import torch
import numpy as np
import pandas as pd


def count_gc(relative_starts, relative_end, gene_ixs, onehot_promoters, window):
    eps = 1e-5
    gc = []
    for relative_start, relative_end, gene_ix in zip(
        relative_starts, relative_end, gene_ixs
    ):
        start_ix = gene_ix * (window[1] - window[0]) + relative_start
        end_ix = gene_ix * (window[1] - window[0]) + relative_end
        gc.append(
            onehot_promoters[start_ix:end_ix, [1, 2]].sum() / (end_ix - start_ix + eps)
        )

    gc = torch.hstack(gc).numpy()[:, None]

    return gc


def count_dinuc(relative_starts, relative_end, gene_ixs, onehot_promoters, window):
    gc = []
    eps = 1e-5
    for relative_start, relative_end, gene_ix in zip(
        relative_starts, relative_end, gene_ixs
    ):
        start_ix = gene_ix * (window[1] - window[0]) + relative_start
        end_ix = gene_ix * (window[1] - window[0]) + relative_end
        nuc = torch.where(onehot_promoters[start_ix:end_ix])[1]
        gc.append(
            torch.bincount(nuc[1:] + (nuc[:-1] * 4), minlength=16)
            / (end_ix - start_ix + eps)
        )

    gc = torch.stack(gc).numpy()

    return gc


def count_motifs(relative_starts, relative_end, gene_ixs, motifscan, window):
    motif_indices = []
    for relative_start, relative_end, gene_ix in zip(
        relative_starts, relative_end, gene_ixs
    ):
        start_ix = gene_ix * (window[1] - window[0]) + relative_start
        end_ix = gene_ix * (window[1] - window[0]) + relative_end
        motif_indices.append(
            motifscan.indices[motifscan.indptr[start_ix] : motifscan.indptr[end_ix]]
        )
    if len(motif_indices) > 0:
        motif_indices = np.hstack(motif_indices)
    else:
        motif_indices = np.array([], dtype=np.int64)
    motif_counts = np.bincount(motif_indices, minlength=motifscan.n_motifs)

    return motif_counts


def count_motifs_genewise(
    relative_starts, relative_end, gene_ixs, motifscan, n_genes, window
):
    motif_indices = []
    for relative_start, relative_end, gene_ix in zip(
        relative_starts, relative_end, gene_ixs
    ):
        start_ix = gene_ix * (window[1] - window[0]) + relative_start
        end_ix = gene_ix * (window[1] - window[0]) + relative_end
        motif_indices.append(
            motifscan.indices[motifscan.indptr[start_ix] : motifscan.indptr[end_ix]]
            + gene_ix * motifscan.n_motifs
        )
    if len(motif_indices) > 0:
        motif_indices = np.hstack(motif_indices)
    else:
        motif_indices = np.array([], dtype=np.int64)
    motif_counts = np.bincount(
        motif_indices, minlength=motifscan.n_motifs * n_genes
    ).reshape((n_genes, motifscan.n_motifs))

    return motif_counts


def count_motifs_slicewise(relative_starts, relative_end, gene_ixs, motifscan, window):
    motif_indices = []
    n_slices = len(relative_starts)
    for relative_start, relative_end, gene_ix, slice_ix in zip(
        relative_starts, relative_end, gene_ixs, range(n_slices)
    ):
        start_ix = gene_ix * (window[1] - window[0]) + relative_start
        end_ix = gene_ix * (window[1] - window[0]) + relative_end
        motif_indices.append(
            motifscan.indices[motifscan.indptr[start_ix] : motifscan.indptr[end_ix]]
            + slice_ix * motifscan.n_motifs
        )
    if len(motif_indices) > 0:
        motif_indices = np.hstack(motif_indices)
    else:
        motif_indices = np.array([], dtype=np.int64)
    motif_counts = np.bincount(
        motif_indices, minlength=motifscan.n_motifs * n_slices
    ).reshape((n_slices, motifscan.n_motifs))

    return motif_counts


def select_background(
    position_slices,
    gene_ixs_slices,
    onehot_promoters,
    window,
    n_genes,
    n_random=100,
    n_background=10,
    seed=None,
):
    window_oi_gc = count_gc(
        position_slices[:, 0],
        position_slices[:, 1],
        gene_ixs_slices,
        onehot_promoters,
        window,
    )
    # window_oi_gc = count_dinuc(
    #     position_slices[:, 0],
    #     position_slices[:, 1],
    #     gene_ixs_slices,
    #     onehot_promoters,
    #     window,
    # )

    if seed is not None:
        rg = np.random.RandomState(seed)
    else:
        rg = np.random.RandomState()

    # random position
    position_slices_repeated = position_slices.repeat(n_random, 0)
    random_position_slices = np.zeros_like(position_slices_repeated)

    random_position_slices[:, 0] = rg.randint(
        np.ones(position_slices_repeated.shape[0]) * window[0],
        np.ones(position_slices_repeated.shape[0]) * window[1]
        - (position_slices_repeated[:, 1] - position_slices_repeated[:, 0])
        + 1,
    )
    random_position_slices[:, 1] = random_position_slices[:, 0] + (
        position_slices_repeated[:, 1] - position_slices_repeated[:, 0]
    )
    # random_position_slices = position_slices_repeated

    # random gene
    random_gene_ixs_slices = rg.randint(n_genes, size=random_position_slices.shape[0])
    # random_gene_ixs_slices = gene_ixs_slices.repeat(n_random, 0)

    window_random_gc = count_gc(
        random_position_slices[:, 0],
        random_position_slices[:, 1],
        random_gene_ixs_slices,
        onehot_promoters,
        window,
    )
    # window_random_gc = count_dinuc(
    #     random_position_slices[:, 0],
    #     random_position_slices[:, 1],
    #     random_gene_ixs_slices,
    #     onehot_promoters,
    #     window,
    # )

    random_difference = np.sqrt(
        (
            (
                window_random_gc.reshape(
                    (position_slices.shape[0], n_random, window_random_gc.shape[-1])
                )
                - window_oi_gc[:, None, :]
            )
            ** 2
        ).mean(-1)
    )

    assert n_random >= n_background

    chosen_background = np.argsort(random_difference, axis=1)[
        :, :n_background
    ].flatten()
    chosen_background_idx = (
        np.repeat(np.arange(position_slices.shape[0]), n_background) * n_random
        + chosen_background
    )

    # control
    # fig, ax = plt.subplots()
    # ax.scatter(window_oi_gc[:, 0], window_random_gc[::n_random, 0])
    # ax.scatter(window_oi_gc[:, 0], window_random_gc[chosen_background_idx][::n_background, 0])
    #

    background_position_slices = random_position_slices[chosen_background_idx]
    background_gene_ixs_slices = random_gene_ixs_slices[chosen_background_idx]

    # control
    # gc_back = count_gc(background_position_slices[:, 0], background_position_slices[:, 1], background_gene_ixs_slices, onehot_promoters)
    # ax.scatter(window_oi_gc[:, 13], gc_back[::n_background, 0])
    #

    # control 2
    # window_oi_dinuc = count_dinuc(position_slices[:, 0], position_slices[:, 1], gene_ixs_slices, onehot_promoters)
    # dinuc_back = count_dinuc(background_position_slices[:, 0], background_position_slices[:, 1], background_gene_ixs_slices, onehot_promoters)
    # ax.scatter(window_oi_dinuc[:, 13], dinuc_back[::n_background][:, 13])

    return background_position_slices, background_gene_ixs_slices


import fisher
import statsmodels.stats.multitest
import scipy.stats

# pip install fisher
def enrich_windows(
    motifscan,
    position_slices,
    gene_ixs_slices,
    n_genes,
    window,
    onehot_promoters=None,
    oi_slices=None,
    background_slices=None,
    n_background=10,
):
    background_position_slices = None
    if background_slices is not None:
        background_position_slices = position_slices[background_slices]
        background_gene_ixs_slices = gene_ixs_slices[background_slices]
    if oi_slices is not None:
        position_slices = position_slices[oi_slices]
        gene_ixs_slices = gene_ixs_slices[oi_slices]

    motif_counts = count_motifs(
        position_slices[:, 0],
        position_slices[:, 1],
        gene_ixs_slices,
        motifscan,
        window=window,
    )
    n_positions = (position_slices[:, 1] - position_slices[:, 0]).sum()
    # motif_counts_slicewise = count_motifs_slicewise(position_slices[:, 0], position_slices[:, 1], gene_ixs_slices, motifscan.indices, motifscan.indptr, motifscan.n_motifs)
    motif_counts_genewise = count_motifs_genewise(
        position_slices[:, 0],
        position_slices[:, 1],
        gene_ixs_slices,
        motifscan,
        n_genes,
        window,
    )
    n_positions_gene = np.bincount(
        gene_ixs_slices,
        weights=position_slices[:, 1] - position_slices[:, 0],
        minlength=n_genes,
    )
    motif_percs_genewise = motif_counts_genewise / (n_positions_gene[:, None] + 1e-5)

    # if no background, use all regions
    if n_background == 0:
        background_position_slices = np.stack(
            [np.repeat(0, n_genes), np.repeat(window[1] - window[0], n_genes)]
        ).T
        background_gene_ixs_slices = np.arange(n_genes)
    elif background_position_slices is None:
        assert (
            onehot_promoters is not None
        ), "Sequences are required for selecting a background"
        background_position_slices, background_gene_ixs_slices = select_background(
            position_slices,
            gene_ixs_slices,
            onehot_promoters,
            seed=1,
            n_genes=n_genes,
            window=window,
            n_random=n_background * 10,
            n_background=n_background,
        )

    background_motif_counts = count_motifs(
        background_position_slices[:, 0],
        background_position_slices[:, 1],
        background_gene_ixs_slices,
        motifscan,
        window=window,
    )
    background_n_positions = (
        background_position_slices[:, 1] - background_position_slices[:, 0]
    ).sum()

    # create contingencies to calculate conditional odds
    contingencies = (
        np.stack(
            [
                np.stack(
                    [
                        background_n_positions - background_motif_counts,
                        background_motif_counts,
                    ]
                ),
                np.stack([n_positions - motif_counts, motif_counts]),
            ]
        )
        .transpose(2, 0, 1)
        .astype(np.uint)
    )

    # if background == all positions, remove the counts from the slices_oi
    if n_background == 0:
        contingencies[:, [0]] = contingencies[:, [0]] - contingencies[:, [1]]

    p_values = fisher.pvalue_npy(
        contingencies[:, 0, 0],
        contingencies[:, 1, 0],
        contingencies[:, 0, 1],
        contingencies[:, 1, 1],
    )[-1]

    n_motifs = np.bincount(motifscan.indices, minlength=motifscan.n_motifs)

    # create motifscores
    motifscores = pd.DataFrame(
        {
            "odds": ((motif_counts + 1) / (n_positions + 1))
            / ((background_motif_counts + 1) / (background_n_positions + 1)),
            "motif": motifscan.motifs.index,
            "in": motif_counts / n_motifs,
            "perc": motif_counts / n_positions,
            "pval": p_values,
            "contingency": [cont for cont in contingencies],
            "perc_gene": [x for x in motif_percs_genewise.T],
        }
    ).set_index("motif")
    motifscores["logodds"] = np.log(motifscores["odds"])
    motifscores["qval"] = statsmodels.stats.multitest.fdrcorrection(
        motifscores["pval"]
    )[-1]

    return motifscores


# pip install fisher
def enrich_windows_genewise(
    motifscan,
    position_slices,
    gene_ixs_slices,
    n_genes,
    window,
    gene_ids,
    oi_slices=None,
):
    if oi_slices is not None:
        position_slices = position_slices[oi_slices]
        gene_ixs_slices = gene_ixs_slices[oi_slices]

    motif_counts = count_motifs_genewise(
        position_slices[:, 0],
        position_slices[:, 1],
        gene_ixs_slices,
        motifscan,
        n_genes,
        window,
    )
    n_positions = np.bincount(
        gene_ixs_slices,
        weights=position_slices[:, 1] - position_slices[:, 0],
        minlength=n_genes,
    )[:, None]

    background_position_slices = np.stack(
        [np.repeat(0, n_genes), np.repeat(window[1] - window[0], n_genes)]
    ).T
    background_gene_ixs_slices = np.arange(n_genes)

    background_motif_counts = count_motifs_genewise(
        background_position_slices[:, 0],
        background_position_slices[:, 1],
        background_gene_ixs_slices,
        motifscan,
        n_genes,
        window,
    )
    background_n_positions = np.bincount(
        background_gene_ixs_slices,
        weights=background_position_slices[:, 1] - background_position_slices[:, 0],
        minlength=n_genes,
    )[:, None]

    # create contingencies to calculate conditional odds
    contingencies = np.stack(
        [
            np.stack(
                [
                    background_n_positions - background_motif_counts,
                    background_motif_counts,
                ],
                -1,
            ),
            np.stack([n_positions - motif_counts, motif_counts], -1),
        ],
        -2,
    ).astype(
        int
    )  # order is: gene, motif, motif yes/no, oi yes/no

    background_motif_counts

    # if background == all positions, remove the counts from the slices_oi
    contingencies[..., [0], :] = contingencies[..., [0], :] - contingencies[..., [1], :]
    assert contingencies.sum() == n_genes * (window[1] - window[0]) * motifscan.n_motifs

    if contingencies.min() < 0:
        raise ValueError("Likely some slices are overlapping!!")

    contingencies = contingencies.astype(np.uint)

    odds_conditional = []
    for cont in contingencies.reshape(-1, 2, 2):
        odds_conditional.append(
            scipy.stats.contingency.odds_ratio(cont, kind="conditional").statistic
        )

    p_values = fisher.pvalue_npy(
        contingencies[..., 0, 0].flatten(),
        contingencies[..., 1, 0].flatten(),
        contingencies[..., 0, 1].flatten(),
        contingencies[..., 1, 1].flatten(),
    )[-1].reshape(contingencies.shape[0:2])

    # create motifscores
    motifscores = pd.DataFrame(
        {
            "odds": odds_conditional,
            "n": background_motif_counts.flatten(),
            "n_found": motif_counts.flatten(),
            "in": (motif_counts / (background_motif_counts + 1e-5)).flatten(),
            "perc": (motif_counts / (n_positions + 1e-5)).flatten(),
            "pval": p_values.flatten(),
        }
    )
    motifscores.index = (
        pd.DataFrame(columns=gene_ids, index=motifscan.motifs.index).unstack().index
    )
    motifscores["logodds"] = np.log(motifscores["odds"])
    motifscores.loc[
        motifscores["n"] > 0, "qval"
    ] = statsmodels.stats.multitest.fdrcorrection(
        motifscores.loc[motifscores["n"] > 0, "pval"]
    )[
        -1
    ]

    return motifscores


def detect_windows(motifscan, position_slices, gene_ixs_slices, gene_ids, window):
    motif_counts = count_motifs_genewise(
        position_slices[:, 0],
        position_slices[:, 1],
        gene_ixs_slices,
        motifscan,
        len(gene_ids),
        window,
    )
    gene_counts = (
        pd.DataFrame(
            {
                "n_positions": position_slices[:, 1] - position_slices[:, 0],
                "gene_ix": gene_ixs_slices,
            }
        )
        .groupby("gene_ix")["n_positions"]
        .sum()
    )
    gene_counts.index = gene_ids[gene_counts.index].values
    gene_counts = gene_counts.reindex(gene_ids, fill_value=0)

    motifscores = (
        pd.DataFrame(
            motif_counts,
            index=gene_ids,
            columns=motifscan.motifs.index,
        )
        .stack()
        .to_frame(name="n_positions")
    )
    motifscores["n_positions"] = gene_counts.loc[
        motifscores.index.get_level_values("gene")
    ].values

    return motifscores


def enrich_cluster_vs_clusters(motifscan, window, regions, clustering_id, n_genes):
    motifscores = []
    for cluster_id in regions[clustering_id].cat.categories:
        oi_slices = regions[clustering_id] == cluster_id
        background_slices = ~oi_slices
        motifscores_group = enrich_windows(
            motifscan,
            regions[["start", "end"]].values,
            regions["gene_ix"].values,
            oi_slices=oi_slices,
            background_slices=background_slices,
            n_genes=n_genes,
            window=window,
        )
        motifscores_group[clustering_id] = cluster_id

        motifscores.append(motifscores_group)
    motifscores = pd.concat(motifscores).reset_index()
    motifscores = motifscores.reset_index().set_index([clustering_id, "motif"])
    return motifscores


def enrich_cluster_vs_background(
    motifscan,
    window,
    regions,
    clustering_id,
    n_genes,
    onehot_promoters,
    n_background=10,
):
    motifscores = []
    for cluster_id in regions[clustering_id].cat.categories:
        print(cluster_id)
        oi_slices = regions[clustering_id] == cluster_id
        if oi_slices.sum() > 0:
            motifscores_group = enrich_windows(
                motifscan,
                regions[["start", "end"]].values,
                regions["gene_ix"].values,
                oi_slices=oi_slices,
                n_genes=n_genes,
                window=window,
                onehot_promoters=onehot_promoters,
                n_background=n_background,
            )
            motifscores_group[clustering_id] = cluster_id

            motifscores.append(motifscores_group)
    motifscores = pd.concat(motifscores).reset_index()
    motifscores = motifscores.reset_index().set_index([clustering_id, "motif"])
    return motifscores


import tqdm.auto as tqdm


def enrich_cluster_vs_all(motifscan, window, regions, clustering_id, n_genes, gene_ids):
    motifscores = []
    for cluster_id in tqdm.tqdm(regions[clustering_id].cat.categories):
        oi_slices = regions[clustering_id] == cluster_id
        motifscores_group = enrich_windows_genewise(
            motifscan,
            regions[["start", "end"]].values,
            regions["gene_ix"].values,
            oi_slices=oi_slices,
            n_genes=n_genes,
            window=window,
            gene_ids=gene_ids,
        )
        index_names = motifscores_group.index.names
        motifscores_group[clustering_id] = cluster_id

        motifscores.append(motifscores_group)
    motifscores = pd.concat(motifscores).reset_index()
    motifscores = motifscores.reset_index().set_index([clustering_id, *index_names])
    return motifscores



import chromatinhd as chd
import numpy as np
import pandas as pd
import xarray as xr
import pickle
import scipy.stats
import tqdm.auto as tqdm
import numpy as np
import itertools


def zscore(x, dim=0):
    return (x - x.mean(axis=dim, keepdims=True)) / x.std(axis=dim, keepdims=True)


def zscore_relative(x, y, dim=0):
    return (x - y.mean(axis=dim, keepdims=True)) / y.std(axis=dim, keepdims=True)


def fdr(p_vals):
    from scipy.stats import rankdata

    ranked_p_values = rankdata(p_vals)
    fdr = p_vals * len(p_vals) / ranked_p_values
    fdr[fdr > 1] = 1

    return fdr


class GenePairWindow(chd.flow.Flow):
    design = chd.flow.Stored("design")

    genes = chd.flow.Stored("genes", default=set)

    def score(
        self,
        fragments,
        transcriptome,
        models,
        folds,
        genes,
        censorer,
        force=False,
        device="cuda",
    ):
        force_ = force
        design = censorer.design.iloc[1:].copy()
        self.design = design

        pbar = tqdm.tqdm(genes, leave=False)

        for gene in pbar:
            pbar.set_description(gene)
            scores_file = self.get_scoring_path(gene) / "scores.pkl"
            interaction_file = self.get_scoring_path(gene) / "interaction.pkl"

            force = force_
            if not all([file.exists() for file in [scores_file, interaction_file]]):
                force = True

            if force:
                deltacor_folds = []
                copredictivity_folds = []
                lost_folds = []
                for fold, model in zip(folds, models):
                    predicted, expected, n_fragments = model.get_prediction_censored(
                        fragments,
                        transcriptome,
                        censorer,
                        cell_ixs=np.concatenate(
                            [fold["cells_validation"], fold["cells_test"]]
                        ),
                        genes=[gene],
                        device=device,
                    )

                    # select 1st gene, given that we're working with one gene anyway
                    predicted = predicted[..., 0]
                    expected = expected[..., 0]
                    n_fragments = n_fragments[..., 0]

                    # calculate delta cor per cell
                    # calculate effect per cellxgene combination
                    predicted_censored = predicted[1:]
                    predicted_full = predicted[0][None, ...]
                    predicted_full_norm = zscore(predicted_full, 1)
                    predicted_censored_norm = zscore_relative(
                        predicted_censored, predicted_full, 1
                    )

                    expected_norm = zscore(expected[None, ...], 1)

                    celldeltacor = -np.abs(
                        predicted_censored_norm - expected_norm
                    ) - -np.abs(predicted_full_norm - expected_norm)
                    with np.errstate(divide="ignore", invalid="ignore"):
                        copredictivity = np.corrcoef(celldeltacor)
                    copredictivity[np.isnan(copredictivity)] = 0.0

                    copredictivity_folds.append(copredictivity)

                    cor = chd.utils.paircor(predicted, expected, dim=-1)
                    deltacor = cor[1:] - cor[0]

                    lost = (n_fragments[0] - n_fragments[1:]).mean(-1)

                    deltacor_folds.append(deltacor)
                    lost_folds.append(lost)

                lost_folds = np.stack(lost_folds, 0)
                deltacor_folds = np.stack(deltacor_folds, 0)
                copredictivity_folds = np.stack(copredictivity_folds, 0)
                print(copredictivity_folds.shape)

                result = xr.Dataset(
                    {
                        "deltacor": xr.DataArray(
                            deltacor_folds,
                            coords=[
                                ("model", np.arange(len(models))),
                                ("window", design.index),
                            ],
                        ),
                        "lost": xr.DataArray(
                            lost_folds,
                            coords=[
                                ("model", np.arange(len(models))),
                                ("window", design.index),
                            ],
                        ),
                    }
                )

                windows_oi = lost_folds.mean(0) > 1e-3

                interaction = xr.DataArray(
                    copredictivity_folds[:, windows_oi][:, :, windows_oi],
                    coords=[
                        ("model", np.arange(len(models))),
                        ("window1", design.index[windows_oi]),
                        ("window2", design.index[windows_oi]),
                    ],
                )

                pickle.dump(result, scores_file.open("wb"))
                pickle.dump(interaction, interaction_file.open("wb"))

                self.genes = self.genes | {gene}

    def get_plotdata(self, gene):
        """
        Get plotdata for a gene
        """
        interaction = pickle.load(
            open(
                self.get_scoring_path(gene) / "interaction.pkl",
                "rb",
            )
        )

        plotdata = interaction.mean("model").to_dataframe("cor").reset_index()
        plotdata["window1"] = plotdata["window1"].astype("category")
        plotdata["window2"] = plotdata["window2"].astype("category")

        plotdata = (
            pd.DataFrame(
                itertools.combinations(self.design.index, 2),
                columns=["window1", "window2"],
            )
            .set_index(["window1", "window2"])
            .join(plotdata.set_index(["window1", "window2"]))
        )
        plotdata = plotdata.reset_index().fillna({"cor": 0.0})
        plotdata["window_mid1"] = self.design.loc[plotdata["window1"]][
            "window_mid"
        ].values
        plotdata["window_mid2"] = self.design.loc[plotdata["window2"]][
            "window_mid"
        ].values
        plotdata["dist"] = np.abs(plotdata["window_mid1"] - plotdata["window_mid2"])
        plotdata = plotdata.query("(window_mid1 < window_mid2)")
        # plotdata = plotdata.query("dist > 1000")

        # x = interaction.stack({"window1_window2": ["window1", "window2"]}).values
        # print(x.shape)
        # scores_statistical = []
        # for i in range(x.shape[1]):
        #     scores_statistical.append(scipy.stats.ttest_1samp(x[:, i], 0).pvalue)
        # scores_statistical = pd.DataFrame({"pval": scores_statistical})
        # scores_statistical["pval"] = scores_statistical["pval"].fillna(1.0)
        # scores_statistical["qval"] = fdr(scores_statistical["pval"])

        # plotdata["pval"] = scores_statistical["pval"].values
        # plotdata["qval"] = scores_statistical["qval"].values

        plotdata.loc[plotdata["dist"] < 1000, "cor"] = 0.0

        return plotdata

    def get_scoring_path(self, gene):
        path = self.path / f"{gene}"
        path.mkdir(parents=True, exist_ok=True)
        return path




import torch
class SparseDenseAdam(torch.optim.Optimizer):
    def __init__(self, parameters_sparse, parameters_dense, lr = 1e-3, weight_decay = 0., **kwargs):
        if len(parameters_sparse) == 0:
            self.optimizer_sparse = None
        else:
            self.optimizer_sparse = torch.optim.SparseAdam(
                parameters_sparse,
                lr = lr,
                **kwargs
            )
        if len(parameters_dense) == 0:
            self.optimizer_dense = None
        else:
            self.optimizer_dense = torch.optim.Adam(
                parameters_dense,
                lr=lr,
                weight_decay = weight_decay,
                **kwargs
            )
            
    def zero_grad(self):
        if self.optimizer_sparse is not None:
            self.optimizer_sparse.zero_grad()
        if self.optimizer_dense is not None:
            self.optimizer_dense.zero_grad()
            
    def step(self):
        if self.optimizer_sparse is not None:
            self.optimizer_sparse.step()
        if self.optimizer_dense is not None:
            self.optimizer_dense.step()


from chromatinhd.grid.grid import Grid, Panel
import numpy as np
import pandas as pd
import matplotlib as mpl


class Broken(Grid):
    def __init__(self, regions, width, gap=1, height=0.5, *args, **kwargs):
        super().__init__(padding_width=gap, *args, **kwargs)

        regions["width"] = regions["end"] - regions["start"]
        regions["ix"] = np.arange(len(regions))

        assert width > gap * (len(regions) - 1), "Width is not enough to contain gaps"

        self.resolution = regions["width"].sum() / (width - (len(regions) - 1) * gap)

        for region, region_info in regions.iterrows():
            subpanel_width = region_info["width"] / self.resolution
            panel, ax = self.add_right(
                Panel((subpanel_width, height)),
            )

            ax.set_xlim(region_info["start"], region_info["end"])
            ax.set_xticks([])
            ax.set_ylim(0, 1)
            ax.set_yticks([])
            if region_info["ix"] != 0:
                ax.spines.left.set_visible(False)
            if region_info["ix"] != len(regions) - 1:
                ax.spines.right.set_visible(False)
            ax.spines.top.set_visible(False)


def add_slanted_x(ax1, ax2, size=4, **kwargs):
    d = 1.0  # proportion of vertical to horizontal extent of the slanted line
    kwargs = dict(
        marker=[(-1, -d), (1, d)],
        markersize=size,
        linestyle="none",
        mew=1,
        clip_on=False,
        **{"color": "k", "mec": "k", **kwargs},
    )
    ax1.plot([1, 1], [0, 1], transform=ax1.transAxes, **kwargs)
    ax2.plot([0, 0], [0, 1], transform=ax2.transAxes, **kwargs)



import numpy as np


def ecdf(data):
    sorted_data = np.sort(data)
    y = np.arange(1, len(sorted_data) + 1) / len(sorted_data)

    sorted_data = np.vstack([sorted_data - 1e-5, sorted_data]).T.flatten()
    y = np.vstack([np.hstack([[0], y[:-1]]), y]).T.flatten()
    return sorted_data, y


def area_under_ecdf(list1):
    x1, y1 = ecdf(list1)
    area = np.trapz(y1, x1)
    return area


def area_between_ecdfs(list1, list2):
    x1, y1 = ecdf(list1)
    x2, y2 = ecdf(list2)

    combined_x = np.concatenate(
        [[-0.001], np.sort(np.unique(np.concatenate((x1, x2))))]
    )

    y1_interp = np.interp(combined_x, x1, y1, left=0, right=1)
    y2_interp = np.interp(combined_x, x2, y2, left=0, right=1)

    ecdf_diff = y1_interp - y2_interp
    area = np.trapz(ecdf_diff, combined_x)
    return area


def relative_area_between_ecdfs(list1, list2):
    x1, y1 = ecdf(list1)
    area = area_between_ecdfs(list1, list2) / np.trapz(y1, x1)
    return area



import torch
import chromatinhd.loss


class TestLoss:
    def test_loss(self):
        x = torch.stack(
            [
                torch.linspace(0, 10, 100),
                torch.linspace(1, 20, 100),
                torch.linspace(1, 20, 100),
            ],
            dim=0,
        )
        y = x.clone()
        y[0] = y[0] * 5 - 3
        y[1] = y[1] * (-4) + 5

        x2 = chromatinhd.loss.normlinreg(x, y)
        assert (torch.sqrt((x2 - y) ** 2) < 1e-3).all()



import math

import torch
from torch.nn import functional as F

DEFAULT_MIN_BIN_WIDTH = 1e-5
DEFAULT_MIN_BIN_HEIGHT = 1e-5


def unconstrained_quadratic_spline(
    inputs,
    unnormalized_widths,
    unnormalized_heights,
    inverse=False,
    tails="linear",
    min_bin_width=DEFAULT_MIN_BIN_WIDTH,
    min_bin_height=DEFAULT_MIN_BIN_HEIGHT,
):
    inside_interval_mask = (inputs <= 1) & (inputs >= 0)

    outputs = torch.zeros_like(inputs)
    logabsdet = torch.zeros_like(inputs)

    outputs[inside_interval_mask], logabsdet[inside_interval_mask] = quadratic_spline(
        inputs=inputs[inside_interval_mask],
        unnormalized_widths=unnormalized_widths,
        # unnormalized_widths=unnormalized_widths[inside_interval_mask, ...],
        unnormalized_heights=unnormalized_heights[inside_interval_mask, ...],
        inverse=inverse,
        min_bin_width=min_bin_width,
        min_bin_height=min_bin_height,
    )

    return outputs, logabsdet


def calculate_widths(unnormalized_widths, min_bin_width=DEFAULT_MIN_BIN_WIDTH):
    widths = F.softmax(unnormalized_widths, dim=-1)
    widths = (
        min_bin_width + (1 - min_bin_width * unnormalized_widths.shape[-1]) * widths
    )
    return widths


def calculate_heights(
    unnormalized_heights, widths, min_bin_height=DEFAULT_MIN_BIN_HEIGHT, local=True
):
    unnorm_heights_exp = torch.exp(unnormalized_heights)

    min_bin_height = 1e-3

    if local:
        # per feature normalization
        unnormalized_area = torch.sum(
            ((unnorm_heights_exp[..., :-1] + unnorm_heights_exp[..., 1:]) / 2) * widths,
            dim=-1,
            keepdim=True,
        )
        heights = unnorm_heights_exp / unnormalized_area
        heights = min_bin_height + (1 - min_bin_height) * heights
    else:
        # global normalization
        unnormalized_area = torch.sum(
            ((unnorm_heights_exp[..., :-1] + unnorm_heights_exp[..., 1:]) / 2) * widths,
        )
        heights = unnorm_heights_exp * unnorm_heights_exp.shape[-2] / unnormalized_area
        heights = min_bin_height + (1 - min_bin_height) * heights

    # to check
    # normalized_area = torch.sum(
    #     ((heights[..., :-1] + heights[..., 1:]) / 2) * widths,
    #     dim=-1,
    #     keepdim=True,
    # )
    # print(normalized_area.sum())

    return heights


def calculate_bin_left_cdf(heights, widths):
    bin_left_cdf = torch.cumsum(
        ((heights[..., :-1] + heights[..., 1:]) / 2) * widths, dim=-1
    )
    bin_left_cdf[..., -1] = 1.0
    bin_left_cdf = F.pad(bin_left_cdf, pad=(1, 0), mode="constant", value=0.0)
    return bin_left_cdf


def calculate_bin_locations(widths):
    bin_locations = torch.cumsum(widths, dim=-1)
    bin_locations[..., -1] = 1.0
    bin_locations = F.pad(bin_locations, pad=(1, 0), mode="constant", value=0.0)
    return bin_locations


def quadratic_spline(
    inputs,
    unnormalized_widths=None,
    unnormalized_heights=None,
    widths=None,
    heights=None,
    bin_left_cdf=None,
    bin_locations=None,
    inverse=False,
):
    # calculate widths
    if widths is None:
        widths = calculate_widths(unnormalized_widths)

    num_bins = widths.shape[-1]
    if widths.ndim == inputs.ndim:
        widths = widths.expand(inputs.shape[0], -1)

    # calculate heights
    if heights is None:
        heights = calculate_heights(unnormalized_heights, widths)

    if heights.ndim == inputs.ndim:
        heights = heights.expand(inputs.shape[0], -1)

    # calculate bin indices
    if bin_left_cdf is None:
        bin_left_cdf = calculate_bin_left_cdf(heights, widths)

    if bin_locations is None:
        bin_locations = calculate_bin_locations(widths)

    if inverse:
        bin_idx = torch.searchsorted(bin_left_cdf, inputs.unsqueeze(-1)).squeeze(-1) - 1
    else:
        bin_idx = (
            torch.searchsorted(bin_locations, inputs.unsqueeze(-1)).squeeze(-1) - 1
        )

    bin_idx = torch.clamp(bin_idx, 0, num_bins - 1)

    if bin_idx.ndim < inputs.ndim:
        bin_idx = bin_idx.unsqueeze(-1)

    # get bin locations/widths for input values
    input_bin_locations = bin_locations.gather(-1, bin_idx.unsqueeze(-1)).squeeze(-1)
    input_bin_widths = widths.gather(-1, bin_idx.unsqueeze(-1)).squeeze(-1)

    if bin_left_cdf.ndim > bin_idx.ndim:
        input_left_cdf = bin_left_cdf.gather(-1, bin_idx.unsqueeze(-1)).squeeze(-1)
        input_left_heights = heights.gather(-1, bin_idx.unsqueeze(-1)).squeeze(-1)
        input_right_heights = heights.gather(-1, bin_idx.unsqueeze(-1) + 1).squeeze(-1)
    else:
        input_left_cdf = bin_left_cdf.gather(-1, bin_idx)
        input_left_heights = heights.gather(-1, bin_idx)
        input_right_heights = heights.gather(-1, bin_idx + 1)

    # calculate log abs det and output
    a = 0.5 * (input_right_heights - input_left_heights) * input_bin_widths
    b = input_left_heights * input_bin_widths
    c = input_left_cdf

    if inverse:
        c_ = c - inputs
        alpha = torch.clamp((-b + torch.sqrt(b.pow(2) - 4 * a * c_)) / (2 * a), 0, 1)
        # avoids alpha = -inf due to perfectly horizontal curve (i.e. a = 0)
        alpha[alpha.isnan()] = 0.0

        outputs = alpha * input_bin_widths + input_bin_locations
        logabsdet = -torch.log(
            (alpha * (input_right_heights - input_left_heights) + input_left_heights)
        )
    else:
        # due to numerical imprecision, alpha can sometimes fall outside of 0 and 1
        alpha = torch.clamp((inputs - input_bin_locations) / input_bin_widths, 0, 1)

        outputs = a * alpha.pow(2) + b * alpha + c

        logabsdet = torch.log(
            (alpha * (input_right_heights - input_left_heights) + input_left_heights)
        )

    outputs = torch.clamp(outputs, 0, 1)

    return outputs, logabsdet



import torch
import numpy as np
import dataclasses
from functools import cached_property

import chromatinhd


@dataclasses.dataclass
class Data:
    relative_coordinates: torch.Tensor
    local_cluster_ixs: torch.Tensor
    local_clusterxvariant_indptr: torch.Tensor
    cluster_cut_lib: torch.Tensor
    local_variant_to_local_variantxgene_reshaper: torch.Tensor
    variantxgene_to_gene: torch.Tensor
    variantxgene_to_local_gene: torch.Tensor
    variantxgene_ixs: torch.Tensor

    expression: torch.Tensor
    genotypes: torch.Tensor

    window_size: np.ndarray
    variants_oi: np.ndarray
    genes_oi: np.ndarray
    clusters_oi: np.ndarray

    def to(self, device):
        for field_name, field in self.__dataclass_fields__.items():
            if field.type is torch.Tensor:
                self.__setattr__(
                    field_name, self.__getattribute__(field_name).to(device)
                )
        return self

    @property
    def n_clusters(self):
        return len(self.clusters_oi)

    @property
    def n_variants(self):
        return len(self.variants_oi)


class ChunkFragments:
    cellxgene_batch_size: int

    preloaded = False

    out_coordinates: torch.Tensor
    out_genemapping: torch.Tensor
    out_local_cellxgene_ix: torch.Tensor

    n_genes: int

    def __init__(
        self,
        fragments: chromatinhd.data.Fragments,
        variant_positions,
        window_size,
        clusters_oi=None,
        n_fragment_per_cellxgene: int = None,
    ):
        # store fragment data
        self.chunkcoords = fragments.chunkcoords.numpy().astype(np.int64)
        self.relcoords = fragments.relcoords.numpy()
        self.chunk_size = fragments.chunk_size
        self.chunkcoords_indptr = fragments.chunkcoords_indptr.numpy()
        self.clusters = fragments.clusters.numpy()

        # store the library size
        self.cluster_cut_lib = torch.bincount(
            fragments.clusters, minlength=len(fragments.clusters_info)
        )

        # create bounds from chromosome positions
        # the windows will be bounded by these positions
        self.bounds = np.hstack(
            [
                fragments.chromosomes["position_start"].values,
                fragments.chromosomes["position_end"].values[[-1]],
            ]
        )

        self.variant_positions = variant_positions

        # cache upper and lower bounds
        self.variant_upper_bounds = self.bounds[
            np.searchsorted(self.bounds, self.variant_positions)
        ]
        self.variant_lower_bounds = self.bounds[
            np.searchsorted(self.bounds, self.variant_positions) - 1
        ]

        # store window size
        self.window_size = window_size

    def load(self, minibatch):
        if not self.preloaded:
            self.preload()

        # optional filtering based on fragments_oi
        coordinates = self.coordinates
        genemapping = self.genemapping
        cellxgene_indptr = self.cellxgene_indptr

        minibatch.cellxgene_oi = cell_gene_to_cellxgene(
            minibatch.cells_oi, minibatch.genes_oi, self.n_genes
        )

        assert len(minibatch.cellxgene_oi) <= self.cellxgene_batch_size, (
            len(minibatch.cellxgene_oi),
            self.cellxgene_batch_size,
        )
        n_fragments = chromatinhd.loaders.extraction.fragments.extract_fragments(
            minibatch.cellxgene_oi,
            cellxgene_indptr,
            coordinates,
            genemapping,
            self.out_coordinates.numpy(),
            self.out_genemapping.numpy(),
            self.out_local_cellxgene_ix.numpy(),
        )
        if n_fragments > self.fragment_buffer_size:
            raise ValueError("n_fragments is too large for the current buffer size")

        if n_fragments == 0:
            n_fragments = 1
        self.out_coordinates.resize_((n_fragments, 2))
        self.out_genemapping.resize_((n_fragments))
        self.out_local_cellxgene_ix.resize_((n_fragments))

        return Data(
            local_cellxgene_ix=self.out_local_cellxgene_ix,
            coordinates=self.out_coordinates,
            n_fragments=n_fragments,
            genemapping=self.out_genemapping,
            window=self.window,
            n_total_genes=self.n_genes,
            **minibatch.items(),
        )



from .utils import get_git_root, get_output, get_code, save, Unpickler, load
from . import sparse
from . import utils
from . import flow
from . import grid
from . import plot
from . import data
from . import train
from . import loaders
from . import loss
from . import embedding
from . import optim
from . import biomart
from . import models

__all__ = [
    "get_git_root",
    "get_output",
    "get_code",
    "save",
    "Unpickler",
    "load",
    "sparse",
    "utils",
    "flow",
    "data",
    "train",
    "loaders",
    "loss",
    "embedding",
    "optim",
    "grid",
    "biomart",
    "models",
    "plot",
]


import numpy as np
import pandas as pd
import pickle

from chromatinhd.flow import (
    Flow,
    StoredTorchInt32,
    Stored,
    StoredTorchInt64,
    TSV,
    Linked,
)

from chromatinhd.data.regions import Regions

import torch
import math
import pathlib
import typing
import tqdm.auto as tqdm


class RawFragments:
    def __init__(self, file):
        self.file = file


class Fragments(Flow):
    """Fragments centered around a gene window"""

    regions = Linked("regions")
    """regions of the fragments"""

    coordinates = StoredTorchInt64("coordinates")
    """Coordinates of the fragments"""

    mapping = StoredTorchInt64("mapping")
    """Mapping of a fragment to a gene and a cell"""

    cellxgene_indptr = StoredTorchInt64("cellxgene_indptr")
    """Index pointers for each cellxgene combination"""

    regions = Linked("regions")

    def create_cellxgene_indptr(self):
        cellxgene = self.mapping[:, 0] * self.n_genes + self.mapping[:, 1]

        if not (cellxgene.diff() >= 0).all():
            raise ValueError(
                "Fragments should be ordered by cell then gene (ascending)"
            )

        n_cellxgene = self.n_genes * self.n_cells
        cellxgene_indptr = torch.nn.functional.pad(
            torch.cumsum(torch.bincount(cellxgene, minlength=n_cellxgene), 0), (1, 0)
        )
        assert self.coordinates.shape[0] == cellxgene_indptr[-1]
        if not (cellxgene_indptr.diff() >= 0).all():
            raise ValueError(
                "Fragments should be ordered by cell then gene (ascending)"
            )
        self.cellxgene_indptr = cellxgene_indptr

    _genemapping = None

    @property
    def genemapping(self):
        if self._genemapping is None:
            self._genemapping = self.mapping[:, 1].contiguous()
        return self._genemapping

    _cellmapping = None

    @property
    def cellmapping(self):
        if self._cellmapping is None:
            self._cellmapping = self.mapping[:, 0].contiguous()
        return self._cellmapping

    var = TSV("var")
    obs = TSV("obs")

    _n_genes = None

    @property
    def n_genes(self):
        if self._n_genes is None:
            self._n_genes = self.var.shape[0]
        return self._n_genes

    _n_cells = None

    @property
    def n_cells(self):
        if self._n_cells is None:
            self._n_cells = self.obs.shape[0]
        return self._n_cells

    @property
    def local_cellxgene_ix(self):
        return self.cellmapping * self.n_genes + self.genemapping

    def estimate_fragment_per_cellxgene(self):
        return math.ceil(self.coordinates.shape[0] / self.n_cells / self.n_genes * 2)

    def create_cut_data(self):
        cut_coordinates = self.coordinates.flatten()
        cut_coordinates = (cut_coordinates - self.window[0]) / (
            self.window[1] - self.window[0]
        )
        keep_cuts = (cut_coordinates >= 0) & (cut_coordinates <= 1)
        cut_coordinates = cut_coordinates[keep_cuts]

        self.cut_coordinates = cut_coordinates

        self.cut_local_gene_ix = self.genemapping.expand(2, -1).T.flatten()[keep_cuts]
        self.cut_local_cell_ix = self.cellmapping.expand(2, -1).T.flatten()[keep_cuts]

    @property
    def genes_oi_torch(self):
        return torch.from_numpy(self.genes_oi).to(self.coordinates.device)

    @property
    def cells_oi_torch(self):
        return torch.from_numpy(self.genes_oi).to(self.coordinates.device)

    @classmethod
    def from_fragments_tsv(
        cls,
        fragments_file: typing.Union[pathlib.Path, str],
        regions: Regions,
        obs: pd.DataFrame,
        path: typing.Union[pathlib.Path, str],
        overwrite=True,
    ):
        """
        Create a Fragments object from a tsv file

        Parameters:
            fragments_file:
                fragments_file of the tsv file
            path:
                folder in which the fragments object will be created
            regions:
                regions object
        """

        if isinstance(fragments_file, str):
            fragments_file = pathlib.Path(fragments_file)
        if isinstance(path, str):
            path = pathlib.Path(path)
        if not fragments_file.exists():
            raise FileNotFoundError(f"File {fragments_file} does not exist")
        if not overwrite and path.exists():
            raise FileExistsError(f"Folder {path} already exists")
        path.mkdir(parents=True, exist_ok=True)

        # region information
        var = pd.DataFrame(index=regions.coordinates.index)
        var["ix"] = np.arange(var.shape[0])

        n_genes = var.shape[0]

        # cell information
        obs["ix"] = np.arange(obs.shape[0])
        cell_to_cell_ix = obs["ix"].to_dict()

        n_cells = obs.shape[0]

        # load fragments tabix
        import pysam

        fragments_tabix = pysam.TabixFile(str(fragments_file))

        coordinates_raw = []
        mapping_raw = []

        for i, (gene, promoter_info) in tqdm.tqdm(
            enumerate(regions.coordinates.iterrows()),
            total=regions.coordinates.shape[0],
            leave=False,
            desc="Processing fragments",
        ):
            gene_ix = var.loc[gene, "ix"]
            start = max(0, promoter_info["start"])

            fragments_promoter = fragments_tabix.fetch(
                promoter_info["chrom"],
                start,
                promoter_info["end"],
                parser=pysam.asTuple(),
            )

            for fragment in fragments_promoter:
                cell = fragment[3]

                # only store the fragment if the cell is actually of interest
                if cell in cell_to_cell_ix:
                    # add raw data of fragment relative to tss
                    coordinates_raw.append(
                        [
                            (int(fragment[1]) - promoter_info["tss"])
                            * promoter_info["strand"],
                            (int(fragment[2]) - promoter_info["tss"])
                            * promoter_info["strand"],
                        ][:: promoter_info["strand"]]
                    )

                    # add mapping of cell/gene
                    mapping_raw.append([cell_to_cell_ix[fragment[3]], gene_ix])

        coordinates = torch.tensor(np.array(coordinates_raw, dtype=np.int64))
        mapping = torch.tensor(np.array(mapping_raw), dtype=torch.int64)

        # Sort `coordinates` and `mapping` according to `mapping`
        sorted_idx = torch.argsort((mapping[:, 0] * var.shape[0] + mapping[:, 1]))
        mapping = mapping[sorted_idx]
        coordinates = coordinates[sorted_idx]

        return cls.create(
            path=path,
            coordinates=coordinates,
            mapping=mapping,
            regions=regions,
            var=var,
            obs=obs,
        )


class ChunkedFragments(Flow):
    chunk_size = Stored("chunk_size")
    chunkcoords = StoredTorchInt64("chunkcoords")
    chunkcoords_indptr = StoredTorchInt32("chunkcoords_indptr")
    clusters = StoredTorchInt32("clusters")
    relcoords = StoredTorchInt32("relcoords")

    clusters = Stored("clusters")
    clusters_info = Stored("clusters_info")
    chromosomes = Stored("chromosomes")



import os
import pathlib
import pickle
import pandas as pd

from .torch import interpolate_1d
from .numpy import indices_to_indptr, indptr_to_indices
from .ansi import colorcodes
from .testing import repeated_kfold_corrected_t_test
from . import ecdf

__all__ = [
    "get_git_root",
    "get_output",
    "get_code",
    "name_window",
    "paircor",
    "interpolate_1d",
    "indices_to_indptr",
    "indptr_to_indices",
    "colorcodes",
    "repeated_kfold_corrected_t_test",
    "ecdf",
]


def get_git_root(cwd=None):
    """
    Gets the first parent root with a a .git folder
    """
    if cwd is None:
        cwd = os.getcwd()
    # go back until we find the git directory, signifying project root
    while ".git" not in os.listdir(cwd) and os.path.realpath(cwd) != "/":
        cwd = os.path.dirname(cwd)

    return pathlib.Path(cwd)


def get_output():
    return get_git_root() / "output"


def get_code():
    return get_git_root() / "code"


def name_window(window_info):
    return (
        window_info["chrom" if "chrom" in window_info.index else "chr"]
        + ":"
        + str(window_info["start"])
        + "-"
        + str(window_info["end"])
    )


def paircor(x, y, dim=-2):
    import torch
    import pandas as pd
    import numpy as np

    if isinstance(x, pd.DataFrame):
        x_ = x.values
        y_ = y.values
        divisor = y_.std(dim) * x_.std(dim)
        divisor[np.isclose(divisor, 0)] = 1.0
        cor = (
            (x_ - x_.mean(dim, keepdims=True)) * (y_ - y_.mean(dim, keepdims=True))
        ).mean(dim) / divisor
        cor = pd.Series(cor, x.index if dim == 1 else x.columns)
    elif torch.is_tensor(x):
        cor = ((x - x.mean(dim, keepdim=True)) * (y - y.mean(dim, keepdim=True))).mean(
            dim
        ) / (y.std(dim) * x.std(dim))
    else:
        divisor = y.std(dim) * x.std(dim)
        divisor[np.isclose(divisor, 0)] = 1.0
        cor = (
            (x - x.mean(dim, keepdims=True)) * (y - y.mean(dim, keepdims=True))
        ).mean(dim) / divisor
    return cor


def paircorr(x, y, dim=-2):
    import numpy as np

    divisor = y.std(dim, keepdims=True) * x.std(dim, keepdims=True)
    divisor[np.isclose(divisor, 0)] = 1.0
    cor = (x - x.mean(dim, keepdims=True)) * (
        y - y.mean(dim, keepdims=True)
    )  # / divisor
    return cor


def paircos(x, y, dim=0):
    import torch
    import pandas as pd
    import numpy as np

    if isinstance(x, pd.DataFrame):
        x_ = x.values
        y_ = y.values
        divisor = (y_**2).sum(dim) * (x_**2).sum(dim)
        divisor[np.isclose(divisor, 0)] = 1.0

        dot = (x_ * y_).sum(dim)
        cos = dot / divisor
        cos = pd.Series(cos, x.index if dim == 1 else x.columns)
    elif torch.is_tensor(x):
        cos = (x * y).sum(dim) / (
            torch.sqrt((y**2).sum(dim)) * torch.sqrt((x**2).sum(dim))
        )
    else:
        divisor = np.sqrt((y**2).sum(dim)) * np.sqrt((x**2).sum(dim))
        divisor[divisor == 0] = 1.0
        cos = (x * y).sum(dim) / divisor
    return cos


def fix_class(obj):
    import importlib

    module = importlib.import_module(obj.__class__.__module__)
    cls = getattr(module, obj.__class__.__name__)
    try:
        obj.__class__ = cls
    except TypeError:
        pass


class Pickler(pickle.Pickler):
    def reducer_override(self, obj):
        if any(
            obj.__class__.__module__.startswith(module) for module in ["chromatinhd."]
        ):
            fix_class(obj)
        else:
            # For any other object, fallback to usual reduction
            return NotImplemented

        return NotImplemented


def save(obj, fh, pickler=None, **kwargs):
    if pickler is None:
        pickler = Pickler
    return pickler(fh).dump(obj)


def crossing(*dfs):
    dfs = [df.copy() if isinstance(df, pd.DataFrame) else df.to_frame() for df in dfs]
    for df in dfs:
        df["___key"] = 0
    if len(dfs) == 0:
        return pd.DataFrame()
    dfs = [df for df in dfs if df.shape[0] > 0]  # remove empty dfs
    base = dfs[0]
    for df in dfs[1:]:
        base = pd.merge(base, df, on="___key")
    return base.drop(columns=["___key"])


class Unpickler(pickle.Unpickler):
    def find_class(self, module, name):
        if module.startswith("peakfreeatac"):
            module = module.replace("peakfreeatac", "chromatinhd")
        return super().find_class(module, name)


def load(file):
    return Unpickler(file).load()



import torch


def interpolate_1d(x: torch.Tensor, xp: torch.Tensor, fp: torch.Tensor) -> torch.Tensor:
    a = (fp[..., 1:] - fp[..., :-1]) / (xp[..., 1:] - xp[..., :-1])
    b = fp[..., :-1] - (a.mul(xp[..., :-1]))

    indices = torch.searchsorted(xp.contiguous(), x.contiguous(), right=False) - 1
    indices = torch.clamp(indices, 0, a.shape[-1] - 1)

    slope = a.index_select(a.ndim - 1, indices)
    intercept = b.index_select(a.ndim - 1, indices)
    return x * slope + intercept


def interpolate_0d(x: torch.Tensor, xp: torch.Tensor, fp: torch.Tensor) -> torch.Tensor:
    a = (fp[..., 1:] - fp[..., :-1]) / (xp[..., 1:] - xp[..., :-1])
    b = fp[..., :-1] - (a.mul(xp[..., :-1]))

    indices = torch.searchsorted(xp.contiguous(), x.contiguous(), right=False) - 1
    indices = torch.clamp(indices, 0, a.shape[-1] - 1)

    return b.index_select(a.ndim - 1, indices)


def indices_to_indptr(x, n):
    return torch.nn.functional.pad(
        torch.cumsum(torch.bincount(x, minlength=n), 0), (1, 0)
    )


def ind2ptr(x, minlength):
    return torch.nn.functional.pad(
        torch.cumsum(torch.bincount(x, minlength=minlength), 0), (1, 0)
    )


def ptr2ind(x):
    return torch.repeat_interleave(torch.arange(len(x) - 1), torch.diff(x))



import torch
import numpy as np
import math
import tqdm.auto as tqdm
import torch_scatter

from chromatinhd.embedding import EmbeddingTensor
from . import spline
from chromatinhd.models import HybridModel

from chromatinhd.flow import Flow, Stored

from chromatinhd.models.diff.loader.minibatches import Minibatcher
from chromatinhd.models.diff.loader.clustering_cuts import (
    ClusteringCuts,
)
from chromatinhd.models.diff.trainer import Trainer
from chromatinhd.loaders import LoaderPool2
from chromatinhd.optim import SparseDenseAdam

import pickle


class Decoder(torch.nn.Module):
    def __init__(
        self,
        n_latent,
        n_genes,
        n_output_components,
        n_layers=1,
        n_hidden_dimensions=32,
        dropout_rate=0.2,
    ):
        super().__init__()

        layers = []
        for i in range(n_layers):
            layers.append(
                torch.nn.Linear(
                    n_hidden_dimensions if i > 0 else n_latent, n_hidden_dimensions
                )
            )
            layers.append(torch.nn.BatchNorm1d(n_hidden_dimensions))
            layers.append(torch.nn.ReLU())
            if dropout_rate > 0.0:
                layers.append(torch.nn.Dropout(0.2))
        self.nn = torch.nn.Sequential(*layers)

        self.n_hidden_dimensions = n_hidden_dimensions
        self.n_output_components = n_output_components

        self.logit_weight = EmbeddingTensor(
            n_genes,
            (n_hidden_dimensions if n_layers > 0 else n_latent, n_output_components),
            sparse=True,
        )
        stdv = 1.0 / math.sqrt(self.logit_weight.weight.size(1))
        # self.logit_weight.weight.data.uniform_(-stdv, stdv)
        self.logit_weight.weight.data.zero_()

        self.rho_weight = EmbeddingTensor(
            n_genes, (n_hidden_dimensions if n_layers > 0 else n_latent,), sparse=True
        )
        stdv = 1.0 / math.sqrt(self.rho_weight.weight.size(1))
        self.rho_weight.weight.data.uniform_(-stdv, stdv)

    def forward(self, latent, genes_oi):
        # genes oi is only used to get the logits
        # we calculate the rho for all genes because we pool using softmax later
        logit_weight = self.logit_weight(genes_oi)
        rho_weight = self.rho_weight.get_full_weight()
        nn_output = self.nn(latent)

        # nn_output is broadcasted across genes and across components
        logit = torch.matmul(nn_output.unsqueeze(1).unsqueeze(2), logit_weight).squeeze(
            -2
        )

        # nn_output has to be broadcasted across genes
        rho = torch.matmul(nn_output.unsqueeze(1), rho_weight.T).squeeze(-2)

        return logit, rho

    def parameters_dense(self):
        return [*self.nn.parameters(), self.rho_weight.weight]

    def parameters_sparse(self):
        return [self.logit_weight.weight]


class BaselineDecoder(torch.nn.Module):
    def __init__(
        self, n_latent, n_genes, n_output_components, n_layers=1, n_hidden_dimensions=32
    ):
        super().__init__()

        layers = []
        for i in range(n_layers):
            layers.append(
                torch.nn.Linear(
                    n_hidden_dimensions if i > 0 else n_latent, n_hidden_dimensions
                )
            )
            layers.append(torch.nn.BatchNorm1d(n_hidden_dimensions))
            layers.append(torch.nn.ReLU())
            layers.append(torch.nn.Dropout(0.2))
        self.nn = torch.nn.Sequential(*layers)

        self.n_hidden_dimensions = n_hidden_dimensions
        self.n_output_components = n_output_components

        self.rho_weight = EmbeddingTensor(
            n_genes, (n_hidden_dimensions if n_layers > 0 else n_latent,), sparse=True
        )
        stdv = 1.0 / math.sqrt(self.rho_weight.weight.size(1)) / 100
        self.rho_weight.weight.data.uniform_(-stdv, stdv)

    def forward(self, latent, genes_oi):
        rho_weight = self.rho_weight(genes_oi)
        nn_output = self.nn(latent)

        # nn_output is broadcasted across genes and across components
        logit = torch.zeros(
            (latent.shape[0], len(genes_oi), self.n_output_components),
            device=latent.device,
        )

        rho = torch.matmul(nn_output, rho_weight.T).squeeze(-1)
        return logit, rho

    def parameters_dense(self):
        return [*self.nn.parameters()]

    def parameters_sparse(self):
        return [self.rho_weight.weight]


class Model(torch.nn.Module, HybridModel):
    def __init__(
        self,
        fragments,
        clustering,
        nbins=(
            128,
            64,
            32,
        ),
        decoder_n_layers=0,
        baseline=False,
        mixture_delta_p_scale_free=False,
        scale_likelihood=False,
        rho_delta_p_scale_free=False,
        mixture_delta_p_scale_dist="normal",
        mixture_delta_p_scale=1.0,
    ):
        super().__init__()

        self.n_total_genes = fragments.n_genes

        self.n_clusters = clustering.n_clusters

        from .spline import DifferentialQuadraticSplineStack, TransformedDistribution

        transform = DifferentialQuadraticSplineStack(
            nbins=nbins,
            n_genes=fragments.n_genes,
        )
        self.mixture = TransformedDistribution(transform)
        n_delta_mixture_components = sum(transform.split_deltas)

        if not baseline:
            self.decoder = Decoder(
                self.n_clusters,
                fragments.n_genes,
                n_delta_mixture_components,
                n_layers=decoder_n_layers,
            )
        else:
            self.decoder = BaselineDecoder(
                self.n_clusters,
                fragments.n_genes,
                n_delta_mixture_components,
                n_layers=decoder_n_layers,
            )

        libsize = torch.bincount(fragments.mapping[:, 0], minlength=fragments.n_cells)
        rho_bias = (
            torch.bincount(fragments.mapping[:, 1], minlength=fragments.n_genes)
            / fragments.n_cells
            / libsize.to(torch.float).mean()
        )
        min_rho_bias = 1e-5
        rho_bias = min_rho_bias + (1 - min_rho_bias) * rho_bias

        self.register_buffer("libsize", libsize)
        self.register_buffer("rho_bias", rho_bias)

        self.track = {}

        if mixture_delta_p_scale_free:
            self.mixture_delta_p_scale = torch.nn.Parameter(
                torch.tensor(math.log(mixture_delta_p_scale), requires_grad=True)
            )
        else:
            self.register_buffer(
                "mixture_delta_p_scale", torch.tensor(math.log(mixture_delta_p_scale))
            )

        self.n_total_cells = fragments.n_cells

        self.scale_likelihood = scale_likelihood

        if rho_delta_p_scale_free:
            self.rho_delta_p_scale = torch.nn.Parameter(
                torch.log(torch.tensor(0.1, requires_grad=True))
            )
        else:
            self.register_buffer("rho_delta_p_scale", torch.tensor(math.log(1.0)))

        self.mixture_delta_p_scale_dist = mixture_delta_p_scale_dist

    def forward_(
        self,
        coordinates,
        clustering,
        genes_oi,
        local_cellxgene_ix,
        localcellxgene_ix,
        local_gene_ix,
    ):
        # decode
        mixture_delta, rho_delta = self.decoder(clustering.to(torch.float), genes_oi)

        # rho
        rho = torch.nn.functional.softmax(torch.log(self.rho_bias) + rho_delta, -1)
        rho_cuts = rho.flatten()[localcellxgene_ix]

        # rho delta kl
        rho_delta_p = torch.distributions.Normal(0.0, torch.exp(self.rho_delta_p_scale))
        rho_delta_kl = rho_delta_p.log_prob(self.decoder.rho_weight(genes_oi))

        # fragment counts
        mixture_delta_cellxgene = mixture_delta.view(
            np.prod(mixture_delta.shape[:2]), mixture_delta.shape[-1]
        )
        mixture_delta = mixture_delta_cellxgene[local_cellxgene_ix]

        self.track["likelihood_mixture"] = likelihood_mixture = self.mixture.log_prob(
            coordinates, genes_oi, local_gene_ix, mixture_delta
        )

        self.track["likelihood_overall"] = likelihood_overall = torch.log(
            rho_cuts
        ) + math.log(self.n_total_genes)

        # overall likelihood
        likelihood = self.track["likelihood"] = likelihood_mixture + likelihood_overall
        likelihood_scale = 1.0

        # mixture kl
        mixture_delta_p = torch.distributions.Normal(
            0.0, torch.exp(self.mixture_delta_p_scale)
        )
        mixture_delta_kl = mixture_delta_p.log_prob(self.decoder.logit_weight(genes_oi))

        # ELBO
        elbo = (
            -likelihood.sum() * likelihood_scale
            - mixture_delta_kl.sum()
            - rho_delta_kl.sum()
        )

        return elbo

    def forward(self, data):
        return self.forward_(
            coordinates=data.cuts.coordinates,
            clustering=data.clustering.onehot,
            genes_oi=data.minibatch.genes_oi_torch,
            local_gene_ix=data.cuts.local_gene_ix,
            local_cellxgene_ix=data.cuts.local_cellxgene_ix,
            localcellxgene_ix=data.cuts.localcellxgene_ix,
        )

    def train_model(self, fragments, clustering, fold, device="cuda", n_epochs=30):
        # set up minibatchers and loaders
        minibatcher_train = Minibatcher(
            fold["cells_train"],
            range(fragments.n_genes),
            n_genes_step=500,
            n_cells_step=200,
        )
        minibatcher_validation = Minibatcher(
            fold["cells_validation"],
            range(fragments.n_genes),
            n_genes_step=10,
            n_cells_step=10000,
            permute_cells=False,
            permute_genes=False,
        )

        loaders_train = LoaderPool2(
            ClusteringCuts,
            dict(
                clustering=clustering,
                fragments=fragments,
                cellxgene_batch_size=minibatcher_train.cellxgene_batch_size,
            ),
            n_workers=10,
        )
        loaders_validation = LoaderPool2(
            ClusteringCuts,
            dict(
                clustering=clustering,
                fragments=fragments,
                cellxgene_batch_size=minibatcher_validation.cellxgene_batch_size,
            ),
            n_workers=5,
        )

        trainer = Trainer(
            self,
            loaders_train,
            loaders_validation,
            minibatcher_train,
            minibatcher_validation,
            SparseDenseAdam(
                self.parameters_sparse(),
                self.parameters_dense(),
                lr=1e-2,
                weight_decay=1e-5,
            ),
            n_epochs=n_epochs,
            checkpoint_every_epoch=1,
            optimize_every_step=1,
            device=device,
        )

        trainer.train()

    def _get_likelihood_cell_gene(
        self, likelihood, local_cellxgene_ix, n_cells, n_genes
    ):
        return torch_scatter.segment_sum_coo(
            likelihood, local_cellxgene_ix, dim_size=n_cells * n_genes
        ).reshape((n_cells, n_genes))

    def get_prediction(
        self,
        fragments,
        clustering,
        cells=None,
        cell_ixs=None,
        genes=None,
        gene_ixs=None,
        device="cuda",
        return_raw=False,
    ):
        """
        Returns the prediction of a dataset
        """
        if cell_ixs is None:
            if cells is None:
                cells = fragments.obs.index
            fragments.obs["ix"] = np.arange(len(fragments.obs))
            cell_ixs = fragments.obs.loc[cells]["ix"].values
        if cells is None:
            cells = fragments.obs.index[cell_ixs]

        if gene_ixs is None:
            if genes is None:
                genes = fragments.var.index
            fragments.var["ix"] = np.arange(len(fragments.var))
            gene_ixs = fragments.var.loc[genes]["ix"].values
        if genes is None:
            genes = fragments.var.index[gene_ixs]

        minibatches = Minibatcher(
            cell_ixs,
            gene_ixs,
            n_genes_step=500,
            n_cells_step=200,
            use_all_cells=True,
            use_all_genes=True,
            permute_cells=False,
            permute_genes=False,
        )
        loaders = LoaderPool2(
            ClusteringCuts,
            dict(
                clustering=clustering,
                fragments=fragments,
                cellxgene_batch_size=minibatches.cellxgene_batch_size,
            ),
            n_workers=5,
        )
        loaders.initialize(minibatches)

        likelihood_mixture = np.zeros((len(cell_ixs), len(gene_ixs)))
        likelihood_overall = np.zeros((len(cell_ixs), len(gene_ixs)))

        cell_mapping = np.zeros(fragments.n_cells, dtype=np.int64)
        cell_mapping[cell_ixs] = np.arange(len(cell_ixs))

        gene_mapping = np.zeros(fragments.n_genes, dtype=np.int64)
        gene_mapping[gene_ixs] = np.arange(len(gene_ixs))

        device = "cuda"
        self.eval()
        self = self.to(device)

        for data in loaders:
            data = data.to(device)
            with torch.no_grad():
                self.forward(data)

            likelihood_mixture[
                np.ix_(
                    cell_mapping[data.minibatch.cells_oi],
                    gene_mapping[data.minibatch.genes_oi],
                )
            ] += (
                self._get_likelihood_cell_gene(
                    self.track["likelihood_mixture"],
                    data.cuts.local_cellxgene_ix,
                    data.minibatch.n_cells,
                    data.minibatch.n_genes,
                )
                .cpu()
                .numpy()
            )
            likelihood_overall[
                np.ix_(
                    cell_mapping[data.minibatch.cells_oi],
                    gene_mapping[data.minibatch.genes_oi],
                )
            ] += (
                self._get_likelihood_cell_gene(
                    self.track["likelihood_overall"],
                    data.cuts.local_cellxgene_ix,
                    data.minibatch.n_cells,
                    data.minibatch.n_genes,
                )
                .cpu()
                .numpy()
            )

        self = self.to("cpu")

        if return_raw:
            return predicted, expected, n_fragments

        import xarray as xr

        result = xr.Dataset(
            {
                "likelihood_mixture": xr.DataArray(
                    likelihood_mixture,
                    dims=("cell", "gene"),
                    coords={"cell": cells, "gene": fragments.var.index},
                ),
                "likelihood_overall": xr.DataArray(
                    likelihood_overall,
                    dims=("cell", "gene"),
                    coords={"cell": cells, "gene": fragments.var.index},
                ),
            }
        )
        return result

    def evaluate_pseudo(self, coordinates, clustering=None, gene_oi=None, gene_ix=None):
        from chromatinhd.models.diff.loader.cuts import Result as CutsResult
        from chromatinhd.models.diff.loader.minibatches import Minibatch
        from chromatinhd.models.diff.loader.clustering import Result as ClusteringResult
        from chromatinhd.models.diff.loader.clustering_cuts import (
            Result as ClusteringCutsResult,
        )

        device = coordinates.device
        if not torch.is_tensor(clustering):
            if clustering is None:
                clustering = 0.0
            clustering = torch.ones((1, self.n_clusters), device=device) * clustering

            print(clustering)

        cells_oi = torch.ones((1,), dtype=torch.long)

        local_cellxgene_ix = torch.tensor([], dtype=torch.long)
        if gene_ix is None:
            if gene_oi is None:
                gene_oi = 0
            genes_oi = torch.tensor([gene_oi], device=device, dtype=torch.long)
            local_gene_ix = torch.zeros_like(coordinates).to(torch.long)
            local_cellxgene_ix = torch.zeros_like(coordinates).to(torch.long)
            localcellxgene_ix = torch.ones_like(coordinates).to(torch.long) * gene_oi
        else:
            assert len(gene_ix) == len(coordinates)
            genes_oi = torch.unique(gene_ix)

            local_gene_mapping = torch.zeros(
                genes_oi.max() + 1, device=device, dtype=torch.long
            )
            local_gene_mapping.index_add_(
                0, genes_oi, torch.arange(len(genes_oi), device=device)
            )

            local_gene_ix = local_gene_mapping[gene_ix]
            local_cell_ix = torch.arange(
                clustering.shape[0], device=local_gene_ix.device
            )
            local_cellxgene_ix = local_cell_ix * len(genes_oi) + local_gene_ix
            localcellxgene_ix = local_cell_ix * self.n_total_genes + gene_ix

        data = ClusteringCutsResult(
            cuts=CutsResult(
                coordinates=coordinates,
                local_cellxgene_ix=local_cellxgene_ix,
                localcellxgene_ix=localcellxgene_ix,
                n_genes=len(genes_oi),
            ),
            clustering=ClusteringResult(
                onehot=clustering,
            ),
            minibatch=Minibatch(
                cells_oi=cells_oi.numpy(),
                genes_oi=genes_oi.numpy(),
            ),
        )

        with torch.no_grad():
            self.forward(data)

        prob = self.track["likelihood"].detach().cpu()
        return prob.detach().cpu()

    def rank(
        self,
        window: np.ndarray,
        n_latent: int,
        device: torch.DeviceObjType = "cuda",
        how: str = "probs_diff_masked",
        prob_cutoff: float = None,
    ) -> np.ndarray:
        n_genes = self.rho_bias.shape[0]

        import pandas as pd
        from chromatinhd.utils import crossing

        self = self.to(device).eval()

        # create design for inference
        design_gene = pd.DataFrame({"gene_ix": np.arange(n_genes)})
        design_latent = pd.DataFrame({"active_latent": np.arange(n_latent)})
        design_coord = pd.DataFrame(
            {"coord": np.arange(window[0], window[1] + 1, step=25)}
        )
        design = crossing(design_gene, design_latent, design_coord)
        batch_size = 100000
        design["batch"] = np.floor(np.arange(design.shape[0]) / batch_size).astype(int)

        # infer
        probs = []
        for _, design_subset in tqdm.tqdm(design.groupby("batch")):
            pseudocoordinates = torch.from_numpy(design_subset["coord"].values).to(
                device
            )
            pseudocoordinates = (pseudocoordinates - window[0]) / (
                window[1] - window[0]
            )
            pseudolatent = torch.nn.functional.one_hot(
                torch.from_numpy(design_subset["active_latent"].values).to(device),
                n_latent,
            ).to(torch.float)
            gene_ix = torch.from_numpy(design_subset["gene_ix"].values).to(device)

            prob = self.evaluate_pseudo(
                pseudocoordinates.to(device),
                latent=pseudolatent.to(device),
                gene_ix=gene_ix,
            )

            probs.append(prob)
        probs = np.hstack(probs)

        probs = probs.reshape(
            (design_gene.shape[0], design_latent.shape[0], design_coord.shape[0])
        )

        # calculate the score we're gonna use: how much does the likelihood of a cut in a window change compared to the "mean"?
        probs_diff = probs - probs.mean(-2, keepdims=True)

        # apply a mask to regions with very low likelihood of a cut
        if prob_cutoff is None:
            prob_cutoff = np.log(1.0)
        mask = probs >= prob_cutoff

        probs_diff_masked = probs_diff.copy()
        probs_diff_masked[~mask] = -np.inf

        ## Single base-pair resolution
        # interpolate the scoring from above but now at single base pairs
        # we may have to smooth this in the future, particularly for very detailed models that already look at base pair resolution
        x = (design["coord"].values).reshape(
            (design_gene.shape[0], design_latent.shape[0], design_coord.shape[0])
        )

        def interpolate(
            x: torch.Tensor, xp: torch.Tensor, fp: torch.Tensor
        ) -> torch.Tensor:
            a = (fp[..., 1:] - fp[..., :-1]) / (xp[..., 1:] - xp[..., :-1])
            b = fp[..., :-1] - (a.mul(xp[..., :-1]))

            indices = (
                torch.searchsorted(xp.contiguous(), x.contiguous(), right=False) - 1
            )
            indices = torch.clamp(indices, 0, a.shape[-1] - 1)
            slope = a.index_select(a.ndim - 1, indices)
            intercept = b.index_select(a.ndim - 1, indices)
            return x * slope + intercept

        desired_x = torch.arange(*window)

        probs_diff_interpolated = interpolate(
            desired_x, torch.from_numpy(x)[0][0], torch.from_numpy(probs_diff)
        ).numpy()
        probs_interpolated = interpolate(
            desired_x, torch.from_numpy(x)[0][0], torch.from_numpy(probs)
        ).numpy()

        if how == "probs_diff_masked":
            # again apply a mask
            probs_diff_interpolated_masked = probs_diff_interpolated.copy()
            mask_interpolated = probs_interpolated >= prob_cutoff
            probs_diff_interpolated_masked[~mask_interpolated] = -np.inf

            return probs_diff_interpolated_masked
        else:
            raise NotImplementedError()


class Models(Flow):
    n_models = Stored("n_models")

    @property
    def models_path(self):
        path = self.path / "models"
        path.mkdir(exist_ok=True)
        return path

    def train_models(
        self, fragments, clustering, folds, device="cuda", n_epochs=30, **kwargs
    ):
        self.n_models = len(folds)
        for fold_ix, fold in [(fold_ix, fold) for fold_ix, fold in enumerate(folds)]:
            desired_outputs = [self.models_path / ("model_" + str(fold_ix) + ".pkl")]
            force = False
            if not all([desired_output.exists() for desired_output in desired_outputs]):
                force = True

            if force:
                model = Model(fragments, clustering, **kwargs)
                model.train_model(
                    fragments, clustering, fold, device=device, n_epochs=n_epochs
                )

                model = model.to("cpu")

                pickle.dump(
                    model,
                    open(self.models_path / ("model_" + str(fold_ix) + ".pkl"), "wb"),
                )

    def __getitem__(self, ix):
        return pickle.load(
            (self.models_path / ("model_" + str(ix) + ".pkl")).open("rb")
        )

    def __len__(self):
        return self.n_models

    def __iter__(self):
        for ix in range(len(self)):
            yield self[ix]



import chromatinhd.grid
import chromatinhd.plot
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np


class Copredictivity(chromatinhd.grid.Panel):
    def __init__(self, plotdata, width):
        super().__init__((width, width / 2))

        norm = mpl.colors.CenteredNorm(0, np.abs(plotdata["cor"]).max())
        cmap = mpl.cm.RdBu_r

        chromatinhd.plot.matshow45(
            self.ax,
            plotdata.set_index(["window_mid1", "window_mid2"])["cor"],
            cmap=cmap,
            norm=norm,
            radius=50,
        )
        self.ax.invert_yaxis()

        panel_copredictivity_legend = self.add_inset(
            chromatinhd.grid.Panel((0.05, 0.8)), pos=(0.0, 0.0), offset=(0.0, 0.2)
        )
        plt.colorbar(
            mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
            cax=panel_copredictivity_legend.ax,
            orientation="vertical",
        )
        panel_copredictivity_legend.ax.set_ylabel(
            "Co-predictivity\n(cor $\\Delta$cor)",
            rotation=0,
            ha="right",
            va="center",
        )
        panel_copredictivity_legend.ax.yaxis.set_ticks_position("left")
        panel_copredictivity_legend.ax.yaxis.set_label_position("left")

    @classmethod
    def from_genepairwindow(cls, genepairwindow, gene, width):
        plotdata = genepairwindow.get_plotdata(gene).reset_index()
        return cls(plotdata, width)



import requests
import pandas as pd
import io
from .cache import cache
import xml.etree.cElementTree as ET


def get_datasets(
    mart="ENSEMBL_MART_ENSEMBL", baseurl="http://www.ensembl.org/biomart/martservice?"
):
    url = f"{baseurl}type=datasets&requestid=biomaRt&mart={mart}"
    if url in cache:
        attributes = cache[url]
    else:
        response = requests.get(url)
        datasets = pd.read_table(
            io.StringIO(response.text),
            sep="\t",
            header=None,
            names=[
                "_",
                "dataset",
                "description",
                "version",
                "assembly",
                "__",
                "___",
                "____",
                "last_update",
            ],
        )
        cache[url] = datasets
    return datasets


class Attribute:
    def __init__(self, name, **kwargs):
        self.name = name
        self.kwargs = kwargs

    def to_xml(self):
        return ET.Element("Attribute", name=self.name, **self.kwargs)


class Filter:
    def __init__(self, name, **kwargs):
        self.name = name
        self.kwargs = kwargs

    def to_xml(self):
        return ET.Element("Filter", name=self.name, **self.kwargs)


class Dataset:
    def __init__(
        self,
        name="hsapiens_gene_ensembl",
        baseurl="http://www.ensembl.org/biomart/martservice?",
        mart="ENSEMBL_MART_ENSEMBL",
    ):
        self.name = name
        self.mart = mart
        self.baseurl = baseurl

    def list_attributes(self):
        url = f"{self.baseurl}type=attributes&dataset={self.name}&mart={self.mart}"

        if url in cache:
            attributes = cache[url]
        else:
            response = requests.get(url)
            attributes = pd.read_table(
                io.StringIO(response.text),
                sep="\t",
                header=None,
                names=[
                    "attribute",
                    "name",
                    "description",
                    "type",
                    "format",
                    "full_attribute",
                    "alternative_attribute",
                ],
            ).set_index("attribute")
            cache[url] = attributes
        return attributes

    def list_filters(self):
        url = f"{self.baseurl}type=filters&dataset={self.name}&mart={self.mart}"

        if url in cache:
            filters = cache[url]
        else:
            response = requests.get(url)
            filters = pd.read_table(
                io.StringIO(response.text),
                sep="\t",
                header=None,
                names=[
                    "filter",
                    "name",
                    "options",
                    "description",
                    "_",
                    "type",
                    "operation",
                    "full_filter",
                    "alternative_filter",
                ],
            ).set_index("filter")
            cache[url] = filters
        return filters

    def attribute(self, name, **kwargs):
        return Attribute(name, **kwargs)

    def filter(self, name, **kwargs):
        return Filter(name, **kwargs)

    def get(self, attributes=[], filters=[]):
        xml = ET.Element(
            "Query",
            virtualSchemaName="default",
            formatter="TSV",
            header="0",
            uniqueRows="0",
            datasetConfigVersion="0.6",
            count="",
        )
        dataset = ET.SubElement(xml, "Dataset", name=self.name, interface="default")
        for filter in filters:
            dataset.append(filter.to_xml())
        for attribute in attributes:
            dataset.append(attribute.to_xml())
        query = ET.tostring(xml).decode("utf-8")
        query = query.replace("\t", "").replace("\n", "")
        url = f"{self.baseurl}query={query}"

        if url in cache:
            result = cache[url]
        else:
            response = requests.get(url)
            result = pd.read_table(
                io.StringIO(response.text),
                sep="\t",
                names=[attribute.name for attribute in attributes],
            )
            cache[url] = result
        return result

    @classmethod
    def from_genome(self, genome):
        if genome in ["mm10", "GRCm38"]:
            return Dataset(
                "mmusculus_gene_ensembl",
                "https://nov2020.archive.ensembl.org/biomart/martservice?",
                "ENSEMBL_MART_ENSEMBL",
            )
        elif genome in ["hg19", "GRCh37"]:
            return Dataset(
                "hsapiens_gene_ensembl",
                "http://grch37.ensembl.org/biomart/martservice?",
                "ENSEMBL_MART_ENSEMBL",
            )
        elif genome in ["hg38", "GRCh38"]:
            return Dataset(
                "hsapiens_gene_ensembl",
                "http://www.ensembl.org/biomart/martservice?",
                "ENSEMBL_MART_ENSEMBL",
            )
        elif genome in ["GRCm39"]:
            return Dataset(
                "mmusculus_gene_ensembl",
                "https://nov2020.archive.ensembl.org/biomart/martservice?",
                "ENSEMBL_MART_ENSEMBL",
            )
        else:
            raise ValueError("Genome not supported")



import torch
import numpy as np


class EmbeddingTensor(torch.nn.Embedding):
    """
    Simple wrapper around torch.nn.Embedding which allows an embedding of any dimensions

    The actual tensor underyling this embedding can be accessed using
    `x.weight`, although this will have dimensions [num_embeddings, prod(embedding_dims)]

    To get the underlying tensor in the correct dimensions, you can use
    `x.get_full_weight()`, which will have dimensions [num_embeddings, *embedding_dims]

    To set the value of this parameter, without tracking gradients, use `x.data`
    """

    def __init__(self, num_embeddings, embedding_dims, *args, **kwargs):
        if not isinstance(embedding_dims, tuple):
            embedding_dims = tuple(embedding_dims)
        if len(embedding_dims) == 0:
            embedding_dim = 1
            embedding_dims = tuple([1])
        else:
            embedding_dim = np.prod(embedding_dims)
        super().__init__(num_embeddings, embedding_dim, *args, **kwargs)
        self.embedding_dims = embedding_dims

    def forward(self, input):
        return super().forward(input).view((input.shape[0], *self.embedding_dims))

    def extra_repr(self) -> str:
        s = "{num_embeddings}, {embedding_dims}"
        if self.padding_idx is not None:
            s += ", padding_idx={padding_idx}"
        if self.max_norm is not None:
            s += ", max_norm={max_norm}"
        if self.norm_type != 2:
            s += ", norm_type={norm_type}"
        if self.scale_grad_by_freq is not False:
            s += ", scale_grad_by_freq={scale_grad_by_freq}"
        if self.sparse is not False:
            s += ", sparse=True"
        return s.format(**self.__dict__)

    def get_full_weight(self):
        return self.weight.view((self.weight.shape[0], *self.embedding_dims))

    @property
    def data(self):
        """
        The data of the parameter in dimensions [num_embeddings, *embedding_dims]
        """
        return self.get_full_weight().data

    @data.setter
    def data(self, value):
        assert value.shape == self.weight.shape
        self.weight.data = value

    @property
    def shape(self):
        """
        The shape of the parameter, i.e. [num_embeddings, *embedding_dims]
        """
        return (self.weight.shape[0], *self.embedding_dims)

    def __getitem__(self, k):
        return self.forward(k)



import chromatinhd.grid
import matplotlib as mpl
from chromatinhd.grid.broken import Broken, Panel
import numpy as np


class Predictivity(chromatinhd.grid.Panel):
    def __init__(self, plotdata, window, width, show_accessibility=False):
        super().__init__((width, 0.5))

        plotdata["effect_sign"] = np.sign(plotdata["effect"])
        plotdata["segment"] = plotdata["effect_sign"].diff().ne(0).cumsum()

        ax = self.ax
        ax.set_xlim(*window)

        for segment, segment_data in plotdata.groupby("segment"):
            color = "tomato" if segment_data["effect"].iloc[0] > 0 else "#0074D9"
            ax.plot(
                segment_data["position"],
                segment_data["deltacor"],
                lw=1,
                color=color,
            )
            ax.fill_between(
                segment_data["position"],
                segment_data["deltacor"],
                0,
                alpha=0.2,
                lw=0,
                color=color,
            )

        # ax.plot(
        #     plotdata["position"],
        #     plotdata["deltacor"],
        #     color="#333",
        #     lw=1,
        # )
        # ax.fill_between(
        #     plotdata["position"],
        #     plotdata["deltacor"],
        #     0,
        #     color="#333",
        #     alpha=0.2,
        #     lw=0,
        # )

        ax.set_ylabel(
            "Predictivity\n($\\Delta$ cor)",
            rotation=0,
            ha="right",
            va="center",
        )

        ax.set_xticks([])
        ax.invert_yaxis()
        ax.set_ylim(0, max(-0.05, ax.get_ylim()[1]))

        if show_accessibility:
            ax2 = self.add_twinx()
            ax2.plot(
                plotdata["position"],
                plotdata["lost"],
                color="tomato",
                # color="#333",
                lw=1,
            )
            ax2.fill_between(
                plotdata["position"],
                plotdata["lost"],
                0,
                color="tomato",
                alpha=0.2,
                lw=0,
            )
            ax2.set_xlim(ax.get_xlim())
            ax2.set_ylabel(
                "# fragments\nper 1kb\nper 1k cells",
                rotation=0,
                ha="left",
                va="center",
                color="tomato",
            )
            ax2.tick_params(axis="y", colors="tomato")
            ax2.set_ylim(
                0,
                plotdata["lost"].max()
                / (plotdata["deltacor"].min() / ax.get_ylim()[1]),
            )

        # change vertical alignment of last y tick to bottom
        ax.set_yticks([0, ax.get_ylim()[1]])
        ax.get_yticklabels()[-1].set_verticalalignment("top")
        ax.get_yticklabels()[0].set_verticalalignment("bottom")

        # vline at tss
        ax.axvline(0, color="#888888", lw=0.5, zorder=-1, dashes=(2, 2))

    @classmethod
    def from_genemultiwindow(
        cls, genemultiwindow, gene, width, show_accessibility=False
    ):
        plotdata = genemultiwindow.get_plotdata(gene).reset_index()
        window = np.array([plotdata["position"].min(), plotdata["position"].max()])
        return cls(plotdata, window, width, show_accessibility=show_accessibility)


class Pileup(chromatinhd.grid.Panel):
    def __init__(self, plotdata, window, width):
        super().__init__((width, 0.5))

        ax = self.ax
        ax.set_xlim(*window)
        ax.plot(
            plotdata["position"],
            plotdata["lost"],
            color="#333",
            lw=1,
        )
        ax.fill_between(
            plotdata["position"],
            plotdata["lost"],
            0,
            color="#333",
            alpha=0.2,
            lw=0,
        )
        ax.set_xlim(ax.get_xlim())
        ax.set_ylabel(
            "# fragments\nper 1kb\nper 1k cells",
            rotation=0,
            ha="right",
            va="center",
        )

        # change vertical alignment of last y tick to bottom
        ax.set_yticks([0, ax.get_ylim()[1]])
        ax.get_yticklabels()[-1].set_verticalalignment("top")
        ax.get_yticklabels()[0].set_verticalalignment("bottom")

        # vline at tss
        ax.axvline(0, color="#888888", lw=0.5, zorder=-1, dashes=(2, 2))

        ax.set_xticks([])
        ax.set_ylim(0)

    @classmethod
    def from_genemultiwindow(cls, genemultiwindow, gene, width):
        plotdata = genemultiwindow.get_plotdata(gene).reset_index()
        window = np.array([plotdata["position"].min(), plotdata["position"].max()])
        return cls(plotdata, window, width)


class PredictivityBroken(Broken):
    def __init__(
        self, plotdata, regions, width, *args, gap=1, break_size=4, height=0.5, **kwargs
    ):
        super().__init__(
            regions=regions,
            height=height,
            width=width,
            gap=gap,
            *args,
            **kwargs,
        )

        ylim = plotdata["deltacor"].min() * 1.05

        for ((region, region_info), (panel, ax)) in zip(
            regions.iterrows(), self.elements[0]
        ):
            plotdata_region = plotdata[
                (plotdata["position"] >= region_info["start"])
                & (plotdata["position"] <= region_info["end"])
            ]

            ax.plot(
                plotdata_region["position"],
                plotdata_region["deltacor"],
                color="#333",
                lw=1,
            )

            c = [
                "tomato" if effect > 0 else "#333"
                for effect in plotdata_region["effect"]
            ]
            ax.scatter(
                plotdata_region["position"],
                plotdata_region["deltacor"],
                c=c,
                s=3,
                zorder=5,
            )
            ax.set_ylim(0, ylim)
            ax.axvline(0, color="#888888", lw=0.5, zorder=-1, dashes=(2, 2))

        ax = self[0, 0].ax
        ax.set_ylabel(
            "$\\Delta$ cor",
            rotation=0,
            ha="right",
            va="center",
        )


class LabelBroken(Broken):
    def __init__(self, regions, *args, **kwargs):
        super().__init__(regions=regions, height=0.00001, *args, **kwargs)

        assert len(self.elements[0]) == len(regions)

        for ((region, region_info), (panel, ax)) in zip(
            regions.iterrows(), self.elements[0]
        ):
            ax.set_xlim(region_info["start"], region_info["end"])
            ax.set_xticks([])
            ax.set_ylim(0, 0.1)
            ax.set_yticks([])
            ax.spines.left.set_visible(False)
            ax.spines.top.set_visible(False)
            ax.spines.right.set_visible(False)

            # plot mean position
            if panel.dim[0] > 0.1:
                ax.set_xticks(
                    [
                        region_info["start"]
                        + (region_info["end"] - region_info["start"]) / 2
                    ]
                )
                ax.tick_params(
                    axis="x",
                    rotation=90,
                    pad=5,
                    length=0,
                    which="major",
                    top=True,
                    labeltop=True,
                    bottom=False,
                    labelbottom=False,
                    labelsize=8,
                )
                ax.xaxis.set_major_formatter(
                    mpl.ticker.FuncFormatter(
                        lambda x, pos: f"{x / 1000:+.0f}kb"
                        if x % 1000 == 0
                        else f"{x / 1000:+.1f}kb"
                    )
                )
            # set region boundaries
            ax.xaxis.set_minor_locator(
                mpl.ticker.FixedLocator([region_info["start"], region_info["end"]])
            )
            ax.tick_params(axis="x", which="minor", top=True, bottom=False)



import numpy as np


def indices_to_indptr(x, n):
    return np.pad(np.cumsum(np.bincount(x, minlength=n), 0), (1, 0))


def indptr_to_indices(x):
    n = len(x) - 1
    return np.repeat(np.arange(n), np.diff(x))



#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION
#cython: language_level=3
cimport cython
import numpy as np
cimport numpy as np
np.import_array()

INT64 = np.int64
ctypedef np.int64_t INT64_t
FLOAT64 = np.float64
ctypedef np.float64_t FLOAT64_t

from libc.stdlib cimport abs as c_abs

# cdef means here that this function is a plain C function (so faster).
# To get all the benefits, we type the arguments and the return value.
cdef INT64_t clip(INT64_t a, INT64_t min_value, INT64_t max_value) nogil:
    return min(max(a, min_value), max_value)

@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
def extract_all(
    INT64_t [:,::1] coordinates,
    INT64_t [::1] genemapping,
    INT64_t [::1] motifscores_indptr,
    INT64_t [::1] motifscores_indices,
    FLOAT64_t [::1] motifscores_data,
    INT64_t window_left,
    INT64_t window_right,
    INT64_t window_width,
    INT64_t cutwindow_left,
    INT64_t cutwindow_right,
    INT64_t [::1] out_fragment_indptr,
    INT64_t [::1] out_motif_ix,
    FLOAT64_t [::1] out_score,
    INT64_t [::1] out_distance,
    INT64_t [:,::1] out_motifcounts,
):
    cdef INT64_t fragment_ix, coord_first, coord_second, gene_ix, position, motifscore_ix, slice_left, slice_right, out_ix, local_fragment_ix, local_motif_n
    
    out_ix = 0 # will store where in the output array we are currently
    local_fragment_ix = 0 # will store the current fragment counting from 0
    local_motif_n = 0 # will store how many motifs have been processed
    
    for fragment_ix in range(coordinates.shape[0]):
        coord_first = coordinates[fragment_ix, 0]
        coord_second = coordinates[fragment_ix, 1]
        gene_ix = genemapping[fragment_ix]
        
        # determine the bounds of the genome slice
        slice_left = coord_first + cutwindow_left - window_left
        if slice_left < 0:
            slice_left = 0
        slice_left = slice_left + gene_ix * window_width
        slice_right = coord_first + cutwindow_right - window_left
        if slice_right >= window_width:
            slice_right = window_width - 1
        slice_right = slice_right + gene_ix * window_width
        
        # loop over each position
        for position in range(slice_left, slice_right):

            # extract for each position its motifs
            for motifscore_ix in range(motifscores_indptr[position], motifscores_indptr[position+1]):
                out_motif_ix[out_ix] = motifscores_indices[motifscore_ix]
                out_score[out_ix] = motifscores_data[motifscore_ix]
                out_distance[out_ix] = position
                
                out_motifcounts[local_fragment_ix, motifscores_indices[motifscore_ix]] += 1
                
                out_ix += 1
                local_motif_n += 1
                
        out_fragment_indptr[local_fragment_ix + 1] = local_motif_n
                
        local_fragment_ix += 1

    return out_ix



@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
def extract_motifcounts(
    INT64_t [:,::1] coordinates,
    INT64_t [::1] genemapping,
    INT64_t [::1] motifscores_indptr,
    INT64_t [::1] motifscores_indices,
    FLOAT64_t [::1] motifscores_data,
    INT64_t window_left,
    INT64_t window_right,
    INT64_t window_width,
    INT64_t cutwindow_left,
    INT64_t cutwindow_right,
    INT64_t [:,::1] out_motifcounts,
):
    cdef INT64_t fragment_ix, coord_first, coord_second, gene_ix, position, motifscore_ix, slice_first_left, slice_first_right, slice_first_mid, slice_second_left, slice_second_mid, slice_second_right, local_fragment_ix
    
    local_fragment_ix = 0 # will store the current fragment counting from 0

    with nogil:
        for fragment_ix in range(coordinates.shape[0]):
            coord_first = coordinates[fragment_ix, 0]
            coord_second = coordinates[fragment_ix, 1]
            gene_ix = genemapping[fragment_ix]
            
            # determine the bounds of the genome slice
            slice_first_left = clip(coord_first + cutwindow_left - window_left, 0, window_width - 1) + gene_ix * window_width
            slice_first_mid = clip(coord_first - window_left, 0, window_width - 1) + gene_ix * window_width
            slice_first_right = clip(coord_first + cutwindow_right - window_left, 0, window_width - 1) + gene_ix * window_width

            slice_second_left = clip(coord_second - cutwindow_right - window_left, 0, window_width - 1) + gene_ix * window_width
            slice_second_mid = clip(coord_second - window_left, 0, window_width - 1) + gene_ix * window_width
            slice_second_right = clip(coord_second - cutwindow_left - window_left, 0, window_width - 1) + gene_ix * window_width

            # avoid "within fragment" slices to overlap with "outside fragment" slices
            if slice_first_right > slice_second_mid:
                slice_first_right = slice_second_mid
            if slice_second_left < slice_first_mid:
                slice_second_left = slice_first_mid

            # avoid the first "within fragment" slice to overlap with the second "within fragment" slice
            # in that case slice_first_right takes priority
            if slice_second_left < slice_first_right:
                slice_second_left = slice_first_right
                
            # outside fragments
            for position in range(slice_first_left, slice_first_mid):
                for motifscore_ix in range(motifscores_indptr[position], motifscores_indptr[position+1]):               
                    out_motifcounts[local_fragment_ix, motifscores_indices[motifscore_ix]] += 1
            for position in range(slice_second_mid, slice_second_right):
                for motifscore_ix in range(motifscores_indptr[position], motifscores_indptr[position+1]):               
                    out_motifcounts[local_fragment_ix, motifscores_indices[motifscore_ix]] += 1
            
            # within fragments
            for position in range(slice_first_mid, slice_first_right):
                for motifscore_ix in range(motifscores_indptr[position], motifscores_indptr[position+1]):               
                    out_motifcounts[local_fragment_ix, motifscores_indices[motifscore_ix]] += 1
            for position in range(slice_second_left, slice_second_mid):
                for motifscore_ix in range(motifscores_indptr[position], motifscores_indptr[position+1]):               
                    out_motifcounts[local_fragment_ix, motifscores_indices[motifscore_ix]] += 1

                    
            local_fragment_ix += 1
    return


@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
def extract_dummy(
    INT64_t [:,::1] coordinates,
    INT64_t [::1] genemapping,
    INT64_t [::1] motifscores_indptr,
    INT64_t [::1] motifscores_indices,
    FLOAT64_t [::1] motifscores_data,
    INT64_t window_left,
    INT64_t window_right,
    INT64_t window_width,
    INT64_t cutwindow_left,
    INT64_t cutwindow_right,
    INT64_t [:,::1] out_motifcounts,
):
    cdef INT64_t fragment_ix, coord_first, coord_second, gene_ix, position, motifscore_ix, slice_left, slice_right, out_ix, local_fragment_ix, local_motif_n
    
    local_fragment_ix = 0 # will store the current fragment counting from 0
    local_motif_n = 0 # will store how many motifs have been processed

    return None



@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
def extract_motifcounts_split(
    INT64_t [:,::1] coordinates,
    INT64_t [::1] genemapping,
    INT64_t [::1] motifscores_indptr,
    INT64_t [::1] motifscores_indices,
    FLOAT64_t [::1] motifscores_data,
    INT64_t n_motifs,
    INT64_t window_left,
    INT64_t window_right,
    INT64_t window_width,
    INT64_t cutwindow_left,
    INT64_t cutwindow_right,
    INT64_t [:,::1] out_motifcounts,
):
    cdef INT64_t coord_first, coord_second, gene_ix, slice_first_left, slice_first_right, slice_first_mid, slice_second_left, slice_second_mid, slice_second_right
    cdef int fragment_ix, position, local_fragment_ix, motifscore_ix
    
    local_fragment_ix = 0 # will store the current fragment counting from 0

    with nogil:
        for fragment_ix in range(coordinates.shape[0]):
            coord_first = coordinates[fragment_ix, 0]
            coord_second = coordinates[fragment_ix, 1]
            gene_ix = genemapping[fragment_ix]
            
            # determine the bounds of the genome slice, 0, window_width - 1) + gene_ix * window_width
            slice_first_left = clip(coord_first + cutwindow_left - window_left, 0, window_width - 1) + gene_ix * window_width
            slice_first_mid = clip(coord_first - window_left, 0, window_width - 1) + gene_ix * window_width
            slice_first_right = clip(coord_first + cutwindow_right - window_left, 0, window_width - 1) + gene_ix * window_width
            slice_second_left = clip(coord_second - cutwindow_right - window_left, 0, window_width - 1) + gene_ix * window_width
            slice_second_mid = clip(coord_second - window_left, 0, window_width - 1) + gene_ix * window_width
            slice_second_right = clip(coord_second - cutwindow_left - window_left, 0, window_width - 1) + gene_ix * window_width

            # avoid "within fragment" slices to overlap with "outside fragment" slices
            if slice_first_right > slice_second_mid:
                slice_first_right = slice_second_mid
            if slice_second_left < slice_first_mid:
                slice_second_left = slice_first_mid

            # avoid the first "within fragment" slice to overlap with the second "within fragment" slice
            # in that case slice_first_right takes priority
            if slice_second_left < slice_first_right:
                slice_second_left = slice_first_right
                
            # outside fragments
            for position in range(slice_first_left, slice_first_mid):
                for motifscore_ix in range(motifscores_indptr[position], motifscores_indptr[position+1]):               
                    out_motifcounts[local_fragment_ix, motifscores_indices[motifscore_ix]] += 1
            for position in range(slice_second_mid, slice_second_right):
                for motifscore_ix in range(motifscores_indptr[position], motifscores_indptr[position+1]):               
                    out_motifcounts[local_fragment_ix, motifscores_indices[motifscore_ix]] += 1
            
            # within fragments
            for position in range(slice_first_mid, slice_first_right):
                for motifscore_ix in range(motifscores_indptr[position], motifscores_indptr[position+1]):               
                    out_motifcounts[local_fragment_ix, motifscores_indices[motifscore_ix] + n_motifs] += 1
            for position in range(slice_second_left, slice_second_mid):
                for motifscore_ix in range(motifscores_indptr[position], motifscores_indptr[position+1]):               
                    out_motifcounts[local_fragment_ix, motifscores_indices[motifscore_ix] + n_motifs] += 1

                    
            local_fragment_ix += 1
    return





@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
def extract_motifcounts_multiple(
    INT64_t [:,::1] coordinates,
    INT64_t [::1] genemapping,
    INT64_t [::1] motifscores_indptr,
    INT64_t [::1] motifscores_indices,
    FLOAT64_t [::1] motifscores_data,
    INT64_t n_motifs,
    INT64_t window_left,
    INT64_t window_right,
    INT64_t window_width,
    INT64_t [::1] cutwindows,
    INT64_t [:,::1] out_motifcounts,
):
    cdef INT64_t coord_first, coord_second, gene_ix, slice_left, slice_right, n_cutwindows
    cdef int local_fragment_ix, fragment_ix, cutwindow_ix, motifscore_ix, position
    
    local_fragment_ix = 0 # will store the current fragment counting from 0
    n_cutwindows = cutwindows.shape[0]

    with nogil:
    # if True:
        for fragment_ix in range(coordinates.shape[0]):
            coord_first = coordinates[fragment_ix, 0]
            coord_second = coordinates[fragment_ix, 1]
            gene_ix = genemapping[fragment_ix]

            # left of first cut
            slice_left = coord_first + cutwindows[0] - window_left
            if slice_left < 0:
                slice_left = 0
            if slice_left > window_width:
                slice_left = window_width -1
            slice_left = slice_left + gene_ix * window_width
            for cutwindow_ix in range(n_cutwindows - 1):
                slice_right = coord_first + cutwindows[cutwindow_ix + 1] - window_left
                if slice_right < 0:
                    slice_right = 0
                if slice_right > window_width:
                    slice_right = window_width -1
                slice_right = slice_right + gene_ix * window_width

                for position in range(slice_left, slice_right):
                    for motifscore_ix in range(motifscores_indptr[position], motifscores_indptr[position+1]):               
                        out_motifcounts[local_fragment_ix, motifscores_indices[motifscore_ix] + n_motifs * cutwindow_ix] += 1

                slice_left = slice_right

            # right of second cut
            slice_right = coord_second - cutwindows[0] - window_left
            if slice_right < 0:
                slice_right = 0
            if slice_right > window_width:
                slice_right = window_width -1
            slice_right = slice_right + gene_ix * window_width
            for cutwindow_ix in range(n_cutwindows - 1):
                slice_left = coord_second - cutwindows[cutwindow_ix + 1] - window_left
                if slice_left < 0:
                    slice_left = 0
                if slice_left > window_width:
                    slice_left = window_width -1
                slice_left = slice_left + gene_ix * window_width

                for position in range(slice_left, slice_right):
                    for motifscore_ix in range(motifscores_indptr[position], motifscores_indptr[position+1]):               
                        out_motifcounts[local_fragment_ix, motifscores_indices[motifscore_ix] + n_motifs * cutwindow_ix] += 1

                slice_right = slice_left
                    
            local_fragment_ix += 1
    return




@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
def extract_motifcounts_relative(
    INT64_t [:,::1] coordinates,
    INT64_t [::1] genemapping,
    INT64_t [::1] motifscores_indptr,
    INT64_t [::1] motifscores_indices,
    FLOAT64_t [::1] motifscores_data,
    INT64_t n_motifs,
    INT64_t window_left,
    INT64_t window_right,
    INT64_t window_width,
    INT64_t cutwindow_left,
    INT64_t cutwindow_right,
    INT64_t promoter_width,
    INT64_t [:,::1] out_motifcounts,
):
    cdef INT64_t position, coord_first, coord_second, gene_ix, slice_first_left, slice_first_right, slice_first_mid, slice_second_left, slice_second_mid, slice_second_right
    cdef int motifscore_ix, fragment_ix, local_fragment_ix, in_promoter
    
    local_fragment_ix = 0 # will store the current fragment counting from 0

    with nogil:
        for fragment_ix in range(coordinates.shape[0]):
            coord_first = coordinates[fragment_ix, 0]
            coord_second = coordinates[fragment_ix, 1]
            gene_ix = genemapping[fragment_ix]

            in_promoter = 0
            if (c_abs(coord_first) < promoter_width) or (c_abs(coord_second) < promoter_width):
                in_promoter = 1
            
            # determine the bounds of the genome slice
            slice_first_left = clip(coord_first + cutwindow_left - window_left, 0, window_width - 1) + gene_ix * window_width
            slice_first_mid = clip(coord_first - window_left, 0, window_width - 1) + gene_ix * window_width
            slice_first_right = clip(coord_first + cutwindow_right - window_left, 0, window_width - 1) + gene_ix * window_width

            slice_second_left = clip(coord_second - cutwindow_right - window_left, 0, window_width - 1) + gene_ix * window_width
            slice_second_mid = clip(coord_second - window_left, 0, window_width - 1) + gene_ix * window_width
            slice_second_right = clip(coord_second - cutwindow_left - window_left, 0, window_width - 1) + gene_ix * window_width

            # avoid "within fragment" slices to overlap with "outside fragment" slices
            if slice_first_right > slice_second_mid:
                slice_first_right = slice_second_mid
            if slice_second_left < slice_first_mid:
                slice_second_left = slice_first_mid

            # avoid the first "within fragment" slice to overlap with the second "within fragment" slice
            # in that case slice_first_right takes priority
            if slice_second_left < slice_first_right:
                slice_second_left = slice_first_right
                
            # outside fragments
            for position in range(slice_first_left, slice_first_mid):
                for motifscore_ix in range(motifscores_indptr[position], motifscores_indptr[position+1]):         
                    out_motifcounts[local_fragment_ix, motifscores_indices[motifscore_ix] + n_motifs * in_promoter] += 1
            for position in range(slice_second_mid, slice_second_right):
                for motifscore_ix in range(motifscores_indptr[position], motifscores_indptr[position+1]):          
                    out_motifcounts[local_fragment_ix, motifscores_indices[motifscore_ix] + n_motifs * in_promoter] += 1
            
            # within fragments
            for position in range(slice_first_mid, slice_first_right):
                for motifscore_ix in range(motifscores_indptr[position], motifscores_indptr[position+1]):
                    out_motifcounts[local_fragment_ix, motifscores_indices[motifscore_ix] + n_motifs * in_promoter] += 1
            for position in range(slice_second_left, slice_second_mid):
                for motifscore_ix in range(motifscores_indptr[position], motifscores_indptr[position+1]):
                    out_motifcounts[local_fragment_ix, motifscores_indices[motifscore_ix] + n_motifs * in_promoter] += 1

                    
            local_fragment_ix += 1
    return




import chromatinhd
import chromatinhd.grid
import matplotlib as mpl
import seaborn as sns
import numpy as np
import pandas as pd

from .differential import Differential
from .differential_expression import DifferentialExpression


class MotifsLegend(chromatinhd.grid.Wrap):
    def __init__(self, motifs_oi, cluster_info, width, panel_height, **kwargs):
        super().__init__(ncol=1, **kwargs)

        motifs_shown = set()

        for cluster in cluster_info.index:
            ax = chromatinhd.grid.Ax((width, panel_height))
            self.add(ax)
            ax = ax.ax

            motifs_oi_cluster = motifs_oi.loc[
                [cluster in clusters for clusters in motifs_oi["clusters"]]
            ]

            n = len(motifs_oi_cluster)
            ax.axis("off")
            ax.set_xlim(-0.5, 2)
            ax.set_ylim((n + 4) / 2 - 4, (n + 4) / 2)

            for i, (motif, motif_info) in enumerate(
                motifs_oi_cluster.loc[
                    motifs_oi_cluster.index.difference(motifs_shown)
                ].iterrows()
            ):
                ax.scatter(
                    [0], [i], color=motif_info["color"], marker="|", clip_on=False
                )
                ax.text(
                    0.1,
                    i,
                    s=motif_info["label"],
                    ha="left",
                    va="center",
                    color=motif_info["color"],
                )
            motifs_shown.update(set(motifs_oi_cluster.index))

        self.set_title("Motifs")


class MotifsHighlighting:
    def __init__(self, wrap_differential, motifdata, motifs_oi, cluster_info):
        motifs_oi["indicator"] = range(len(motifs_oi))
        for cluster, ax in zip(cluster_info.index, wrap_differential):
            ax = ax.ax

            # motifs
            motifs_oi_cluster = motifs_oi.loc[
                [cluster in clusters for clusters in motifs_oi["clusters"]]
            ]
            motifdata_cluster = motifdata.loc[
                motifdata["motif"].isin(motifs_oi_cluster.index)
            ]
            texts = []
            for _, z in motifdata_cluster.iterrows():
                ax.axvline(
                    z["position"],
                    # color="grey",
                    color=motifs_oi.loc[z["motif"], "color"],
                    zorder=5,
                )
                ax.scatter(
                    z["position"],
                    ax.get_ylim()[-1],
                    color=motifs_oi.loc[z["motif"], "color"],
                    zorder=100,
                    marker="v",
                    s=100,
                )
                # text = ax.text(
                #     z["position"],
                #     ax.get_ylim()[-1],
                #     s=str(motifs_oi.loc[z["motif"], "indicator"]),
                #     va="top",
                #     ha="center",
                #     zorder=10,
                #     fontsize=6,
                # )
                # text.set_path_effects(
                #     [
                #         mpl.patheffects.Stroke(linewidth=2, foreground="white"),
                #         mpl.patheffects.Normal(),
                #     ]
                # )
                # texts.append(text)
            # import adjustText

            # adjustText.adjust_text(texts, ax=ax, autoalign="x", va="top", ha="center")


class Peaks(chromatinhd.grid.Ax):
    def __init__(
        self,
        peaks,
        peakcallers,
        window,
        width,
        label_methods=True,
        label_rows=True,
        label_methods_side="right",
        row_height=1,
    ):
        super().__init__((width, row_height * len(peakcallers) / 5))

        ax = self.ax
        ax.set_xlim(*window)
        for peakcaller, peaks_peakcaller in peaks.groupby("peakcaller"):
            y = peakcallers.loc[peakcaller, "ix"]

            if len(peaks_peakcaller) == 0:
                continue
            if ("cluster" not in peaks_peakcaller.columns) or pd.isnull(
                peaks_peakcaller["cluster"]
            ).all():
                for _, peak in peaks_peakcaller.iterrows():
                    rect = mpl.patches.Rectangle(
                        (peak["start"], y),
                        peak["end"] - peak["start"],
                        1,
                        fc="#333",
                        lw=0,
                    )
                    ax.add_patch(rect)
                    ax.plot([peak["start"]] * 2, [y, y + 1], color="grey", lw=0.5)
                    ax.plot([peak["end"]] * 2, [y, y + 1], color="grey", lw=0.5)
            else:
                n_clusters = peaks_peakcaller["cluster"].max() + 1
                h = 1 / n_clusters
                for _, peak in peaks_peakcaller.iterrows():
                    rect = mpl.patches.Rectangle(
                        (peak["start"], y + peak["cluster"] / n_clusters),
                        peak["end"] - peak["start"],
                        h,
                        fc="#333",
                        lw=0,
                    )
                    ax.add_patch(rect)
            if y > 0:
                ax.axhline(y, color="#DDD", zorder=10, lw=0.5)

        ax.set_ylim(peakcallers["ix"].max() + 1, 0)
        if label_methods:
            ax.set_yticks(peakcallers["ix"] + 0.5)
            ax.set_yticks(peakcallers["ix"].tolist() + [len(peakcallers)], minor=True)
            ax.set_yticklabels(
                peakcallers["label"],
                fontsize=min(16 * row_height, 10),
                va="center",
            )
            if label_rows:
                ax.set_ylabel("CREs", rotation=0, ha="right", va="center")
            else:
                ax.set_ylabel("")
        else:
            ax.set_yticks([])
            ax.set_ylabel("")
        ax.tick_params(
            axis="y",
            which="major",
            length=0,
            pad=1,
            right=label_methods_side == "right",
            left=not label_methods_side == "left",
        )
        ax.tick_params(
            axis="y",
            which="minor",
            length=1,
            pad=1,
            right=label_methods_side == "right",
            left=not label_methods_side == "left",
        )
        ax.yaxis.tick_right()

        ax.set_xticks([])


class Conservation(chromatinhd.grid.Ax):
    def __init__(self, plotdata_conservation, window, width):
        super().__init__((width, 0.3))

        ax = self.ax
        ax.set_xlim(*window)

        ax.plot(
            plotdata_conservation["position"],
            plotdata_conservation["conservation"],
            color="#333",
            lw=1,
        )

        ax.set_ylim(0, 1)
        ax.set_ylabel("Conservation", rotation=0, ha="right", va="center")
        ax.set_xticks([])


class GC(chromatinhd.grid.Ax):
    def __init__(self, plotdata_gc, window, width):
        super().__init__((width, 0.3))

        ax = self.ax
        ax.set_xlim(*window)

        ax.plot(
            plotdata_gc["position"],
            plotdata_gc["gc"],
            color="#333",
            lw=1,
        )

        ax.set_ylim(0, 1)
        ax.set_ylabel("%GC", rotation=0, ha="right", va="center")
        ax.set_xticks([])


def find_runs(x):
    return np.where((np.diff(np.concatenate([[False], x, [False]])) != 0))[0].reshape(
        (-1, 2)
    )


class CommonUnique:
    def __init__(
        self,
        ax,
        peak_position_chosen_oi,
        region_position_chosen_oi,
        expanded_slice_oi,
        window,
        method_info,
    ):
        # add region and peak unique spans
        trans = mpl.transforms.blended_transform_factory(
            y_transform=ax.transAxes, x_transform=ax.transData
        )
        for start, end in find_runs(
            peak_position_chosen_oi & ~region_position_chosen_oi
        ):
            start = start + expanded_slice_oi["start"] + window[0]
            end = end + expanded_slice_oi["start"] + window[0]

            color = method_info.loc["peak", "color"]

            ax.axvspan(
                start,
                end,
                fc=color,
                alpha=0.3,
                lw=0,
            )
            rect = mpl.patches.Rectangle(
                (start, 1),
                end - start,
                0.2,
                transform=trans,
                fc=color,
                clip_on=False,
                lw=0,
            )
            ax.add_patch(rect)
        for start, end in find_runs(
            peak_position_chosen_oi & region_position_chosen_oi
        ):
            color = method_info.loc["common", "color"]

            start = start + expanded_slice_oi["start"] + window[0]
            end = end + expanded_slice_oi["start"] + window[0]
            ax.axvspan(
                start,
                end,
                fc=color,
                alpha=0.3,
                lw=0,
            )
            rect = mpl.patches.Rectangle(
                (start, 1),
                end - start,
                0.2,
                transform=trans,
                fc=color,
                clip_on=False,
                lw=0,
            )
            ax.add_patch(rect)
        for start, end in find_runs(
            ~peak_position_chosen_oi & region_position_chosen_oi
        ):
            color = method_info.loc["region", "color"]

            start = start + expanded_slice_oi["start"] + window[0]
            end = end + expanded_slice_oi["start"] + window[0]
            ax.axvspan(
                start,
                end,
                fc=color,
                alpha=0.3,
                lw=0,
            )

            rect = mpl.patches.Rectangle(
                (start, 1),
                end - start,
                0.2,
                transform=trans,
                fc=color,
                lw=0,
                clip_on=False,
            )
            ax.add_patch(rect)


class LabelSlice:
    def __init__(self, ax, gene_label, cluster_label, slice_oi, window):
        start = slice_oi["start"] + window[0]
        end = slice_oi["end"] + window[0]

        text = ax.annotate(
            f"$\\it{{{gene_label}}}$ $\\bf{{{cluster_label}}}$",
            (0, 1),
            (2, 2),
            va="bottom",
            ha="left",
            xycoords="axes fraction",
            textcoords="offset points",
            fontsize=6,
            color="#999",
            zorder=200,
        )
        text.set_path_effects(
            [
                mpl.patheffects.Stroke(foreground="#FFFFFFFF", linewidth=2),
                mpl.patheffects.Normal(),
            ]
        )

        trans = mpl.transforms.blended_transform_factory(
            y_transform=ax.transAxes, x_transform=ax.transData
        )
        text = ax.annotate(
            f"{start:+}",
            (start, 1),
            (-2, -2),
            va="top",
            ha="right",
            xycoords=trans,
            textcoords="offset points",
            fontsize=6,
            color="#999",
            zorder=200,
        )
        text.set_path_effects(
            [
                mpl.patheffects.Stroke(foreground="white", linewidth=2),
                mpl.patheffects.Normal(),
            ]
        )
        text = ax.annotate(
            f"{end:+}",
            (end, 1),
            (2, -2),
            va="top",
            ha="left",
            xycoords=trans,
            textcoords="offset points",
            fontsize=6,
            color="#999",
            zorder=200,
        )
        text.set_path_effects(
            [
                mpl.patheffects.Stroke(foreground="white", linewidth=2),
                mpl.patheffects.Normal(),
            ]
        )


class LegendResolution:
    def __init__(self, ax, resolution):
        pass


import io

chromstate_info = pd.read_table(
    io.StringIO(
        """ix	mnemomic	description	color_name	color_code
1	TssA	Active TSS	Red	255,0,0
2	TssAFlnk	Flanking Active TSS	Orange Red	255,69,0
3	TxFlnk	Transcr. at gene 5' and 3'	LimeGreen	50,205,50
4	Tx	Strong transcription	Green	0,128,0
5	TxWk	Weak transcription	DarkGreen	0,100,0
6	EnhG	Genic enhancers	GreenYellow	194,225,5
7	Enh	Enhancers	Yellow	255,255,0
8	ZNF/Rpts	ZNF genes & repeats	Medium Aquamarine	102,205,170
9	Het	Heterochromatin	PaleTurquoise	138,145,208
10	TssBiv	Bivalent/Poised TSS	IndianRed	205,92,92
11	BivFlnk	Flanking Bivalent TSS/Enh	DarkSalmon	233,150,122
12	EnhBiv	Bivalent Enhancer	DarkKhaki	189,183,107
13	ReprPC	Repressed PolyComb	Silver	128,128,128
14	ReprPCWk	Weak Repressed PolyComb	Gainsboro	192,192,192
15	Quies	Quiescent/Low	White	255,255,255
"""
    )
).set_index("mnemomic")
chromstate_info["color"] = [
    np.array(c.split(","), dtype=float) / 255 for c in chromstate_info["color_code"]
]


class Annot(chromatinhd.grid.Ax):
    def __init__(self, plotdata, window, width, cluster_info):
        super().__init__((width, len(cluster_info) * 0.15))

        ax = self.ax
        ax.set_xlim(*window)

        cluster_info["ix"] = np.arange(len(cluster_info))

        for cluster in cluster_info.index:
            plotdata_cluster = plotdata.query("cluster == @cluster")
            y = cluster_info.loc[cluster, "ix"]
            for _, annot in plotdata_cluster.iterrows():
                color = chromstate_info.loc[annot["name"], "color"]
                patch = mpl.patches.Rectangle(
                    (annot["start"], y), annot["end"] - annot["start"], 1, fc=color
                )
                ax.add_patch(patch)

        ax.set_ylim(0, len(cluster_info))
        ax.set_yticks(cluster_info["ix"] + 0.5)
        ax.set_yticklabels(cluster_info.index)
        ax.invert_yaxis()

        # ax.plot(
        #     plotdata["position"],
        #     plotdata["gc"],
        #     color="#333",
        #     lw=1,
        # )

        # ax.set_ylim(0, 1)
        # ax.set_ylabel("%GC", rotation=0, ha="right", va="center")
        # ax.set_xticks([])


class AnnotLegend(chromatinhd.grid.Ax):
    def __init__(self, ax_annot, width=3):
        super().__init__((width, ax_annot.dim[1]))

        ax = self.ax

        chromstate_info["ix"] = np.arange(len(chromstate_info))

        ax.set_ylim(0, self.dim[1])

        for chromstate, chromstate_info_ in chromstate_info.iterrows():
            ax.text(
                0,
                chromstate_info_["ix"] * 0.1,
                fontsize=8,
                s=chromstate_info_["description"],
                color=chromstate_info_["color"],
                va="top",
            )
        ax.invert_yaxis()
        ax.axis("off")



from chromatinhd.flow import Flow, CompressedNumpyInt64, CompressedNumpyFloat64, Stored
from chromatinhd.utils import indptr_to_indices
import pandas as pd


class Motifscan(Flow):
    indptr = CompressedNumpyInt64("indptr")
    indices = CompressedNumpyInt64("indices")
    data = CompressedNumpyFloat64("data")
    shape = Stored("shape")
    n_motifs = Stored("n_motifs")

    _motifs = None

    @property
    def motifs(self):
        if self._motifs is None:
            self._motifs = pd.read_pickle(self.path / "motifs.pkl")
        return self._motifs

    @motifs.setter
    def motifs(self, value):
        value.index.name = "gene"
        value.to_pickle(self.path / "motifs.pkl")
        self._motifs = value


class GWAS(Motifscan):
    association = Stored("association")
    window = Stored("window")

    def get_motifdata(self, promoter: pd.Series):
        assert "ix" in promoter.index
        gene_ix = promoter["ix"]

        indptr_start = gene_ix * (self.window[1] - self.window[0])
        indptr_end = (gene_ix + 1) * (self.window[1] - self.window[0])

        print(indptr_start, indptr_end)

        indptr = self.indptr[indptr_start:indptr_end]
        motif_indices = self.indices[indptr[0] : indptr[-1]]
        position_indices = indptr_to_indices(indptr - indptr[0]) + self.window[0]

        plotdata_snps = pd.DataFrame(
            {
                "position": position_indices,
                "motif": self.motifs.iloc[motif_indices].index.values,
            }
        )
        plotdata_snps = (
            plotdata_snps.groupby("position").agg({"motif": list}).reset_index()
        )

        for position in plotdata_snps["position"]:
            genome_position = (
                promoter["tss"]
                + position * promoter["strand"]
                + 1 * (promoter["strand"] == -1)
            )
            assoc = self.association.loc[
                (self.association["chr"] == promoter.chr)
                & (self.association["start"] == genome_position)
            ]
            if len(assoc) > 0:
                plotdata_snps.loc[
                    plotdata_snps["position"] == position, "rsid"
                ] = assoc["snp"].values[0]
                plotdata_snps.loc[
                    plotdata_snps["position"] == position, "snp_main"
                ] = assoc["snp_main"].values[0]
        return plotdata_snps



import numpy as np
import chromatinhd as chd


class TestClusteringCuts:
    def test_example(self, example_fragments, example_clustering):
        loader = chd.models.diff.loader.ClusteringCuts(
            fragments=example_fragments,
            clustering=example_clustering,
            cellxgene_batch_size=10000,
        )

        minibatch = chd.models.diff.loader.Minibatch(
            cells_oi=np.arange(20), genes_oi=np.arange(5), phase="train"
        )
        result = loader.load(minibatch)



import numpy as np
import pandas as pd
import matplotlib as mpl


def matshow45(ax, series, radius=None, cmap=None, norm=None):

    """
    fig, ax = plt.subplots()
    plotdata = pd.DataFrame(
        np.random.rand(10, 10),
        index=np.arange(10),
        columns=np.arange(10),
    ).stack()
    ax.set_aspect(1)
    matshow45(ax, plotdata)
    """
    offsets = []
    colors = []

    assert len(series.index.names) == 2
    x = series.index.get_level_values(0)
    y = series.index.get_level_values(1)

    if radius is None:
        radius = np.diff(y)[0] / 2

    centerxs = x + (y - x) / 2
    centerys = (y - x) / 2

    xlim = [
        centerxs.min() - radius,
        centerxs.max() + radius,
    ]
    ax.set_xlim(xlim)

    ylim = [
        centerys.unique().min(),
        centerys.unique().max(),
    ]
    ax.set_ylim(ylim)

    if norm is None:
        norm = mpl.colors.Normalize(vmin=series.min(), vmax=series.max())

    if cmap is None:
        cmap = mpl.cm.get_cmap()

    vertices = []

    for centerx, centery, value in zip(centerxs, centerys, series.values):
        center = np.array([centerx, centery])
        offsets.append(center)
        colors.append(cmap(norm(value)))

        vertices.append(
            [
                center + np.array([radius * 1.1, 0]),
                center + np.array([0, radius * 1.1]),
                center + np.array([-radius * 1.1, 0]),
                center + np.array([0, -radius * 1.1]),
            ]
        )
    vertices = np.array(vertices)
    collection = mpl.collections.PolyCollection(
        vertices,
        ec=None,
        lw=0,
        fc=colors,
    )

    ax.add_collection(collection)

    for x in [xlim[1]]:
        x2 = x
        x1 = x2 + (xlim[0] - x2) / 2
        y2 = 0
        y1 = x2 - x1

        if True:
            color = "black"
            lw = 0.8
            zorder = 10
        elif False:
            color = "#eee"
            lw = 0.5
            zorder = -1
        ax.plot(
            [x1 + radius, x2 + radius],
            [y1, y2],
            zorder=zorder,
            color=color,
            lw=lw,
        )

        x1, x2 = (xlim[1] - xlim[0]) / 2 - x2, (xlim[1] - xlim[0]) / 2 - x1
        ax.plot(
            [-x1 - radius, -x2 - radius],
            [y1, y2],
            zorder=zorder,
            color=color,
            lw=lw,
        )
    ax.axis("off")



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.7
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # ChromatinHD-*diff*

# %% tags=["hide_code", "hide_output"]
# autoreload
import IPython

if IPython.get_ipython() is not None:
    IPython.get_ipython().run_line_magic("load_ext", "autoreload")
    IPython.get_ipython().run_line_magic("autoreload", "2")

# %% tags=["hide_output"]
import chromatinhd as chd
import matplotlib.pyplot as plt

# %% [markdown]
# ChromatinHD-<i>pred</i> uses accessibility fragments to predict gene expression. As such, it can detect features such as broad or narrow positioning of fragments, or fragment sizes, that are predictive for gene expression.

# %% [markdown]
# We first load in all the input data:

# %%
import pathlib
dataset_folder = pathlib.Path("example")
fragments = chd.data.Fragments(dataset_folder / "fragments")
transcriptome = chd.data.Transcriptome(dataset_folder / "transcriptome")
folds = chd.data.folds.Folds(dataset_folder / "folds" / "5x1")
clustering = chd.data.Clustering(dataset_folder / "clustering")

# %%
cluster_info = clustering.cluster_info
cluster_info.index.name = "cluster"
clustering.cluster_info = cluster_info

# %% [markdown]
# ## Train the models

# %% [markdown]
# The basic ChromatinHD-*diff* model

# %%
import chromatinhd.models.diff.model.cutnf

# %%
models = chd.models.diff.model.cutnf.Models(dataset_folder / "models" / "cutnf", reset = True)

# %% tags=["hide_output"]
models.train_models(fragments, clustering, folds, device = "cuda", n_epochs = 10, nbins = (256, 128, 64, 32))

# %% [markdown]
# ## Interpret positionally

# %% [markdown]
# Currently, the model is a purely positional model, and as such we can only interpret it positionally.

# %%
import chromatinhd.models.diff.interpret.genepositional

# %%
genepositional = chromatinhd.models.diff.interpret.genepositional.GenePositional(
    path=models.path / "interpret" / "genepositional"
)
genepositional.score(
    fragments,
    clustering,
    models,
    folds,
    genes=transcriptome.gene_id(
        [
            "CCL4",
            "IL1B",
            "EBF1",
        ]
    ),
    force=True,
)

# %%
symbol = "CCL4"

fig = chd.grid.Figure(chd.grid.Grid(padding_height=0.05, padding_width=0.05))
width = 10

region = fragments.regions.coordinates.loc[transcriptome.gene_id(symbol)]
panel_genes = chd.plot.genome.genes.Genes.from_region(region, width=width)
fig.main.add_under(panel_genes)

plotdata, plotdata_mean = genepositional.get_plotdata(transcriptome.gene_id(symbol))
panel_differential = chd.models.diff.plot.Differential(
    plotdata, plotdata_mean, cluster_info = clustering.cluster_info, panel_height = 0.5, width=width
)
fig.main.add_under(panel_differential)

panel_expression = chd.models.diff.plot.DifferentialExpression.from_transcriptome(
    transcriptome = transcriptome, clustering = clustering, gene = transcriptome.gene_id(symbol), panel_height = 0.5
)
fig.main.add_right(panel_expression, row = panel_differential)

fig.plot()

# %%



import numpy as np
import scipy.stats as stats


def autocorrelation(x):
    """
    Calculate the autocorrelation of a sequence.

    Args:
    x (list): A sequence of numbers.

    Returns:
    float: Autocorrelation value.
    """
    n = len(x)
    x_bar = np.mean(x)
    numerator = np.sum([(x[i] - x_bar) * (x[i - 1] - x_bar) for i in range(1, n)])
    denominator = np.sum([(x[i] - x_bar) ** 2 for i in range(n)])

    return numerator / denominator


def repeated_kfold_corrected_t_test(
    performance_A, performance_B, k, num_repeats, alpha=0.05
):
    """
    Perform the corrected t-test between two learning algorithms A and B for repeated K-fold cross-validation.

    Args:
    performance_A (list): A list of performance scores for algorithm A.
    performance_B (list): A list of performance scores for algorithm B.
    k (int): Number of folds in the cross-validation.
    num_repeats (int): Number of times the K-fold cross-validation was repeated.
    alpha (float): Significance level, default is 0.05.

    Returns:
    bool: True if there is a significant difference, False otherwise.
    float: t-statistic value.
    float: p-value.
    """

    if (
        len(performance_A) != len(performance_B)
        or len(performance_A) != k * num_repeats
    ):
        raise ValueError(
            "Performance scores for each algorithm should have the same length and match k * num_repeats."
        )

    n = k * num_repeats

    d = [performance_A[i] - performance_B[i] for i in range(n)]
    d_bar = np.mean(d)
    s_d = np.std(d, ddof=1)

    rho = autocorrelation(d)
    effective_sample_size = n * (1 - rho) / (1 + rho)

    t_statistic = d_bar / (s_d / np.sqrt(effective_sample_size))
    degrees_of_freedom = effective_sample_size - 1
    p_value = 2 * (1 - stats.t.cdf(abs(t_statistic), degrees_of_freedom))

    reject_null_hypothesis = p_value < alpha

    return reject_null_hypothesis, t_statistic, p_value


import scipy.stats


def repeated_kfold_corrected_t_test(diff, r, k, n_train, n_test):
    diff_corrected = 1 / (k * r) * diff.sum()
    variance = 1 / (k * r - 1) * ((diff - diff_corrected) ** 2).sum()
    t = (
        1
        / (k * r)
        * diff_corrected
        / np.sqrt((1 / (k * r) + n_test / n_train) * variance)
    )

    p_value = scipy.stats.t.cdf(t, r * k - 1)
    return t



import torch
import numpy as np
import pyximport

pyximport.install(
    reload_support=True,
    language_level=3,
    setup_args=dict(include_dirs=[np.get_include()]),
)
from . import fragments_helpers  # pylint: disable=C0413,E0611
import dataclasses

import chromatinhd.data.fragments


@dataclasses.dataclass
class Result:
    coordinates: torch.Tensor
    local_cellxgene_ix: torch.Tensor
    localcellxgene_ix: torch.Tensor
    n_genes: int

    @property
    def local_gene_ix(self):
        return self.local_cellxgene_ix % self.n_genes

    @property
    def local_cell_ix(self):
        return torch.div(self.local_cellxgene_ix, self.n_genes, rounding_mode="floor")

    def to(self, device):
        for field_name, field in self.__dataclass_fields__.items():
            if field.type is torch.Tensor:
                self.__setattr__(
                    field_name, self.__getattribute__(field_name).to(device)
                )
        return self


def cell_gene_to_cellxgene(cells_oi, genes_oi, n_genes):
    return (cells_oi[:, None] * n_genes + genes_oi).flatten()


class Cuts:
    """
    Provides cuts data for a minibatch.
    """

    cellxgene_batch_size: int

    preloaded = False

    out_coordinates: torch.Tensor
    out_genemapping: torch.Tensor
    out_local_cellxgene_ix: torch.Tensor

    n_genes: int

    def __init__(
        self,
        fragments: chromatinhd.data.fragments.Fragments,
        cellxgene_batch_size: int,
        n_fragment_per_cellxgene: int = None,
    ):
        self.cellxgene_batch_size = cellxgene_batch_size

        # store auxilliary information
        window = fragments.regions.window
        self.window = window
        self.window_width = window[1] - window[0]

        # store fragment data
        self.cellxgene_indptr = fragments.cellxgene_indptr.numpy()
        self.coordinates = fragments.coordinates.numpy()
        self.genemapping = fragments.genemapping.numpy()

        # create buffers for coordinates
        if n_fragment_per_cellxgene is None:
            n_fragment_per_cellxgene = fragments.estimate_fragment_per_cellxgene()
        fragment_buffer_size = n_fragment_per_cellxgene * cellxgene_batch_size
        self.fragment_buffer_size = fragment_buffer_size

        self.n_genes = fragments.n_genes

    def preload(self):
        self.out_coordinates = torch.from_numpy(
            np.zeros((self.fragment_buffer_size, 2), dtype=np.int64)
        )  # .pin_memory()
        self.out_genemapping = torch.from_numpy(
            np.zeros(self.fragment_buffer_size, dtype=np.int64)
        )  # .pin_memory()
        self.out_local_cellxgene_ix = torch.from_numpy(
            np.zeros(self.fragment_buffer_size, dtype=np.int64)
        )  # .pin_memory()

        self.preloaded = True

    def load(self, minibatch):
        if not self.preloaded:
            self.preload()

        minibatch.cellxgene_oi = cell_gene_to_cellxgene(
            minibatch.cells_oi, minibatch.genes_oi, self.n_genes
        )

        assert len(minibatch.cellxgene_oi) <= self.cellxgene_batch_size, (
            len(minibatch.cellxgene_oi),
            self.cellxgene_batch_size,
        )
        n_fragments = fragments_helpers.extract_fragments(
            minibatch.cellxgene_oi,
            self.cellxgene_indptr,
            self.coordinates,
            self.genemapping,
            self.out_coordinates.numpy(),
            self.out_genemapping.numpy(),
            self.out_local_cellxgene_ix.numpy(),
        )
        if n_fragments > self.fragment_buffer_size:
            raise ValueError("n_fragments is too large for the current buffer size")

        if n_fragments == 0:
            n_fragments = 1
        self.out_coordinates.resize_((n_fragments, 2))
        self.out_genemapping.resize_((n_fragments))
        self.out_local_cellxgene_ix.resize_((n_fragments))

        # convert to cut data
        cut_coordinates = self.out_coordinates.flatten()
        cut_coordinates = (cut_coordinates - self.window[0]) / (
            self.window[1] - self.window[0]
        )
        keep_cuts = (cut_coordinates >= 0) & (cut_coordinates <= 1)
        cut_coordinates = cut_coordinates[keep_cuts]

        local_cellxgene_ix = self.out_local_cellxgene_ix

        local_cellxgene_ix = local_cellxgene_ix.expand(2, -1).T.flatten()[keep_cuts]
        local_cell_ix = torch.div(
            local_cellxgene_ix, self.n_genes, rounding_mode="floor"
        )
        localcellxgene_ix = (
            local_cell_ix * self.n_genes
            + self.out_genemapping.expand(2, -1).T.flatten()[keep_cuts]
        )

        return Result(
            coordinates=cut_coordinates,
            local_cellxgene_ix=local_cellxgene_ix,
            localcellxgene_ix=localcellxgene_ix,
            n_genes=len(minibatch.genes_oi),
        )



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.7
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Data preparation

# %% tags=["hide_code", "hide_output"]
# autoreload
import IPython

if IPython.get_ipython() is not None:
    IPython.get_ipython().run_line_magic("load_ext", "autoreload")
    IPython.get_ipython().run_line_magic("autoreload", "2")

# %% tags=["hide_output"]
import chromatinhd as chd

# %% [markdown]
# To speed up training and inference, ChromatinHD stores several intermediate files to disk. This includes preprocessed data and models. These will be stored in the example folder.

# %%
import pathlib

dataset_folder = pathlib.Path("example")
dataset_folder.mkdir(exist_ok=True)

# %% tags=["hide_code", "hide_output"]
import shutil

for file in dataset_folder.iterdir():
    if file.is_file():
        file.unlink()
    else:
        shutil.rmtree(file)

# %% [markdown]
# For this quickstart, we will use a tiny example dataset extracted from the 10X multiome PBMC example data. We'll copy over both the h5ad for the transcriptomics data, and the fragments.tsv for the accessibility data.

# %%
import pkg_resources
import shutil

DATA_PATH = pathlib.Path(
    pkg_resources.resource_filename("chromatinhd", "data/examples/pbmc10ktiny/")
)

# # copy all files from data path to dataset folder
for file in DATA_PATH.iterdir():
    shutil.copy(file, dataset_folder / file.name)

# %%
# !ls {dataset_folder}

# %% [markdown]
#
# ### Transcriptomics

# %%
import scanpy as sc

adata = sc.read(dataset_folder / "transcriptome.h5ad")

# %%
transcriptome = chd.data.Transcriptome.from_adata(
    adata, path=dataset_folder / "transcriptome"
)

# %%
# !ls {dataset_folder}/*

# %% [markdown]
# <div class="result">
#   <details class="warning">
#   <summary><strong>Batch effects</strong></summary>
#   <p>
#       Currently, none of the ChromatinHD models directly supports batch effects, although this will likely be added in the future. If you have batch effects, the current recommended workflow depends on the source of the batch effect:
#     <ul>
#     <li>If it mainly comes from ambient mRNA, we recommend to use the corrected data. The reason is that this batch effect will likely not be present in the ATAC-seq data.</li>
#     <li>If it mainly comes from biological differences (e.g. cell stress, patient differences, ...), we recommend to use the uncorrected data. The reason is that this batch effect will likely be reflected in the ATAC-seq data as well, given that the genes are truly differentially regulated between the cells.</li>
#     </ul>
#   </p>
#   </details>
# </div>

# %% [markdown]
# ### Clusters

# %% [markdown]
# Although not needed for every model, for interpretation it can be helpful to store some clustering.

# %%
clustering = chd.data.Clustering.from_labels(
    adata.obs["celltype"], path=dataset_folder / "clustering"
)

# %%
# !ls {clustering.path}

# %% [markdown]
# ### Regions of interest

# %% [markdown]
# ChromatinHD defines a set of regions of interest, typically surrounding the one or more transcription start sites of a gene. Given a set of genes from the transcriptomics data and a genome, we can define these regions as follows:

# %% [markdown]
# We first get the canonical transcripts for each gene.

# %%
biomart_dataset = chd.biomart.Dataset.from_genome("GRCh38")
canonical_transcripts = chd.biomart.get_canonical_transcripts(
    biomart_dataset, transcriptome.var.index
)

# %% [markdown]
# Now we can define the regions around the TSS. In this case we choose -10kb and +10kb around a TSS, although in real situations this will typically be much bigger (e.g. -100kb - +100kb)

# %%
regions = chd.data.Regions.from_canonical_transcripts(
    canonical_transcripts,
    path=dataset_folder / "regions",
    window=[-10000, 10000],
)

# %%
# !ls -lh {dataset_folder}/*

# %% [markdown]
# <div class="result">
#   <details class="note">
#   <summary><strong>Gene vs TSS coordinates</strong></summary>
#   <p>The coordinates of the canonical transcript often do not correspond to the gene annotation that are used for e.g. RNA-seq analysis. The reason is that gene coordinates are defined based on the largest transcript in both ends.
#   </p>
#   </details>
# </div>

# %% [markdown]
#
# ### ATAC-seq
#
# ChromatinHD simply requires a `fragments.tsv` file. This contains for each fragment its chromosome, start, end and cell barcode.
#
# - When using Cellranger, this file will be produced by the pipeline.
# - If you have a bam file, [you can use sinto to create the fragment file](https://timoast.github.io/sinto/basic_usage.html)
#

# %% [markdown]
# The fragment file should be indexed.

# %%
if not (dataset_folder / "fragments.tsv.gz.tbi").exists():
    import subprocess

    subprocess.run(
        [
            "tabix",
            dataset_folder / "fragments.tsv.gz",
        ]
    )

# %%
fragments = chd.data.Fragments.from_fragments_tsv(
    dataset_folder / "fragments.tsv.gz",
    regions,
    obs=transcriptome.obs,
    path=dataset_folder / "fragments",
)

# %%
fragments.create_cellxgene_indptr()

# %%
# !ls {dataset_folder}/*

# %% [markdown]
# ### Training folds

# %% [markdown]
# The final set of data are the training folds that will be used to train - and test - the model. For basic models this is simply done by randomly sampling cells.

# %%
folds = chd.data.folds.Folds(dataset_folder / "folds" / "5x1").sample_cells(
    fragments, 5, 1
)

# %%

# %%



import chromatinhd.data.transcriptome
import dataclasses
import torch


@dataclasses.dataclass
class Result:
    value: torch.Tensor

    def to(self, device):
        self.value = self.value.to(device)
        return self


class Transcriptome:
    def __init__(
        self,
        transcriptome: chromatinhd.data.transcriptome.Transcriptome,
        layer: str = "X",
    ):
        self.X = torch.from_numpy(transcriptome.layers[layer])

    def load(self, minibatch):
        X = self.X[minibatch.cells_oi, :][:, minibatch.genes_oi]
        return Result(value=X)



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.7
#   kernelspec:
#     display_name: chromatinhd
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Installation

# %% tags=["hide_code"]
# autoreload
import IPython

if IPython.get_ipython() is not None:
    IPython.get_ipython().run_line_magic("load_ext", "autoreload")
    IPython.get_ipython().run_line_magic("autoreload", "2")

# %% [markdown]
# <pre class="bash">
# # from github
# <b>pip install git+https://github.com/DeplanckeLab/ChromatinHD</b>
#
# # (soon) using pip
# pip install chromatinhd
#
# # (soon) using conda
# conda install -c bioconda chromatinhd
# </pre>

# %% [markdown]
#
# To use the GPU, ensure that a PyTorch version was installed with cuda enabled:
#

# %%
import torch

torch.cuda.is_available()  # should return True
torch.cuda.device_count()  # should be >= 1

# %% [markdown]
#
# If not, follow the instructions at https://pytorch.org/get-started/locally/. You may have to re-install PyTorch.

# %% tags=["hide_output"]
import chromatinhd as chd
